from terraform_compliance.common.error_handling import Error
from terraform_compliance.extensions.ext_radish_bdd import skip_step


def its_key_metadata_has_something(_step_obj, key, value, has_step=True, regex_match=None):
    match = _step_obj.context.match
    match.regex_flag = regex_match

    if match.equals(key, 'values'):
        Error(_step_obj, 'The key "values" is not metadata, please use the "When its property has something" step instead.')
    
    found_list = []
    for resource in _step_obj.context.stash:
        # document this behavior
        key_metadata = match.get(resource, key, default={})
        if not isinstance(key_metadata, (list, dict, set)):
            key_metadata = [key_metadata]

        if has_step and not match.contains(key_metadata, value):
            continue
        elif not has_step and match.contains(key_metadata, value):
            continue

        found_list.append(resource)

    _step_obj.context.stash = found_list

    if not found_list:
        if has_step:
            skip_step(_step_obj, message='Could not find "{}" in "{}" metadata.'.format(value, key))
        if not has_step:
            skip_step(_step_obj, message='Found "{}" in "{}" metadata for all resources.'.format(value, key))


from terraform_compliance.common.defaults import Defaults
from terraform_compliance.common.helper import Match
from terraform_compliance.common.error_handling import Error
import re


def look_for_bdd_tags(_step_obj):
    _step_obj.context.no_failure = False
    _step_obj.context.failure_class = None
    _step_obj.context.no_skip = False
    _step_obj.context.skip_class = None # to pick a tagname that user used
    _step_obj.context.lines_to_noskip = []
    _step_obj.context.bad_tags = False
    _step_obj.context.resources_to_exclude = r''

    defaults = Defaults()

    if hasattr(_step_obj, 'all_tags') and len(_step_obj.all_tags) > 0:
        _step_obj.context.case_insensitivity = True

        for tag in _step_obj.all_tags:
            if tag.name.lower() in defaults.no_failure_tags:
                _step_obj.context.no_failure = True
                _step_obj.context.failure_class = tag.name
            elif tag.name.lower() in defaults.case_sensitive_tags:
                _step_obj.context.case_insensitivity = False
            elif tag.name.lower() in defaults.no_skip_tags:
                _step_obj.context.no_skip = True
                _step_obj.context.skip_class = tag.name
                _step_obj.context.lines_to_noskip = [-1]
            # noskip
            elif re.search(r'({})_at_lines?_.*'.format('|'.join(defaults.no_skip_tags)), tag.name.lower()):
                # check for '@noskip at line x'
                regex = r'({})_at_lines?((_\d*)*)'.format('|'.join(defaults.no_skip_tags))
                
                m = re.search(regex, tag.name.lower())
                if m is not None:
                    _step_obj.context.no_skip = True
                    _step_obj.context.skip_class = tag.name

                    line_numbers_string = m.group(2)

                    try:
                        line_numbers = map(int, line_numbers_string.strip('_').split('_'))
                        if _step_obj.context.lines_to_noskip != [-1]: # no need to worry about this tag if we already have a general noskip
                            _step_obj.context.lines_to_noskip.extend(line_numbers)
                    
                    except Exception as e:
                        Error(_step_obj, f'A tag was determined to be a noskip, but line numbers could not be grouped by the regex {regex}\n{e}')
            # exclude resources
            elif re.search(r'({})_.+\..+'.format('|'.join(defaults.exclude_resources_tags)), tag.name.lower()):
                regex = r'({})_(.+)'.format('|'.join(defaults.exclude_resources_tags))
                
                m = re.search(regex, tag.name.lower())
                if m is not None:
                    resource_address = m.group(2)

                    if _step_obj.context.resources_to_exclude:
                        _step_obj.context.resources_to_exclude += r'|'
                    
                    _step_obj.context.resources_to_exclude += r'({})'.format(resource_address)


    if _step_obj.context.no_failure and _step_obj.context.no_skip:
        _step_obj.context.no_failure = False
        _step_obj.context.bad_tags = True
        Error(_step_obj, f'@{_step_obj.context.failure_class} and @{_step_obj.context.skip_class} tags can not be used at the same time.')

    ## set the match here
    case_sensitive = True if hasattr(_step_obj.context, 'case_insensitivity') and not _step_obj.context.case_insensitivity else False
    _step_obj.context.match = Match(case_sensitive=case_sensitive)

    return _step_obj



# -*- coding: utf-8 -*-

from terraform_compliance.common.helper import (
    seek_key_in_dict,  # importing this purely because the unit tests require it to exist in global scope
    Null
)
from terraform_compliance.common.error_handling import Error


def it_must_contain_something(_step_obj, something, inherited_values=Null, child=False):
    match = _step_obj.context.match
    seek_key_in_dict, seek_regex_key_in_dict_values = match.seek_key_in_dict, match.seek_regex_key_in_dict_values

    prop_list = []

    _step_obj.context.stash = inherited_values if inherited_values is not Null else _step_obj.context.stash

    if _step_obj.context.type in ('resource', 'data'):
        for resource in _step_obj.context.stash:
            if not isinstance(resource, dict) \
                    or 'values' not in resource \
                    or 'address' not in resource \
                    or 'type' not in resource:
                resource = {'values': resource,
                            'address': resource,
                            'type': _step_obj.context.name}

            # not going to use match.get here because the following line is an edge case
            values = resource.get('values', resource.get('expressions', {}))
            if not values:
                values = seek_key_in_dict(resource, something)

            found_values = []
            found_key = Null # this could also become a list
            resource_passed = False
            # set this to True if you get anything from the resource, don't set it to False if you get empty values as there could be other values as well
            if isinstance(values, dict):

                found_key = match.get(values, something, Null)
                if found_key is not Null:
                    found_key = [{something: found_key}]
                else:
                    found_key = seek_key_in_dict(values, something)

                for kv_pair in found_key:
                    # kv_pair must be in {something: found_key} format.
                    if not isinstance(kv_pair, dict):
                        continue # should raise exception
                    # ignore the values that correspond to Null 
                    # Following line could be problematic, how to determine if something is set to be empty or not set? Behavior is provider dependent. 
                    # For now, allow '' and don't allow [] as per user cases.
                    if match.get(kv_pair, something) not in ([],):
                        found_values.append(match.get(kv_pair, something))
                        resource_passed = True

            elif isinstance(values, list):

                for value in values:

                    if isinstance(value, dict):
                        # First search in the keys
                        found_key = seek_key_in_dict(value, something)

                        # The following is an edge case that covers things like aws asg tags (https://www.terraform.io/docs/providers/aws/r/autoscaling_group.html)
                        # Then search in the values with 'key'
                        if not found_key:
                            found_key = seek_regex_key_in_dict_values(value, 'key', something)

                            if found_key:
                                found_key = found_key[0]
                                # not going to use match.get here because the following line is an edge case
                                found_values.extend(value.get('value'))
                                resource_passed = True
                                continue
                    elif isinstance(value, list):
                        _, temp_found_values = it_must_contain_something(_step_obj, something, value, child=True)
                        prop_list.extend(temp_found_values)
                        resource_passed = True

                    elif isinstance(value, (str, bool, int, float)):
                        if match.equals(value, something):
                            found_values.append(value)
                            resource_passed = True

                    if found_key is not Null and len(found_key):

                        for found_key_instance in found_key:
                            if isinstance(found_key_instance, dict):
                                if match.get(found_key_instance, something, Null) not in (Null, [], '', {}):
                                    found_values.append(match.get(found_key_instance, something))
                                    resource_passed = True

            for i, found_val in enumerate(found_values):
                if isinstance(found_val, dict) and 'constant_value' in found_val:
                    found_values[i] = found_val['constant_value']

            for found_val in found_values:
                prop_list.append({'address': resource['address'],
                                'values': found_val,
                                'type': _step_obj.context.name})

            # do not check prop list here because every resource should contain it.
            if not resource_passed and not child: # if nothing was found in this resource, don't error if you're a child
                Error(_step_obj, '{} ({}) does not have {} property.'.format(resource['address'],
                                                                             resource.get('type', ''),
                                                                             something))

        if prop_list:
            _step_obj.context.stash = prop_list
            _step_obj.context.property_name = something
            return something, prop_list

    elif _step_obj.context.type == 'provider':
        prop_list = []
        for provider_data in _step_obj.context.stash:
            values = seek_key_in_dict(provider_data, something)

            if values:
                prop_list.extend(values)
                _step_obj.context.property_name = something
                _step_obj.context.address = '{}.{}'.format(provider_data.get('name', _step_obj.context.addresses),
                                                           provider_data.get('alias', "\b"))

            else:
                Error(_step_obj, '{} {} does not have {} property.'.format(_step_obj.context.addresses,
                                                                           _step_obj.context.type,
                                                                           something))

        if prop_list:
            _step_obj.context.stash = prop_list
            return True

        Error(_step_obj, '{} {} does not have {} property.'.format(_step_obj.context.addresses,
                                                                   _step_obj.context.type,
                                                                   something))


def it_must_not_contain_something(_step_obj, something, inherited_values=Null):
    match = _step_obj.context.match
    seek_key_in_dict, seek_regex_key_in_dict_values = match.seek_key_in_dict, match.seek_regex_key_in_dict_values

    prop_list = []

    _step_obj.context.stash = inherited_values if inherited_values is not Null else _step_obj.context.stash

    if _step_obj.context.type in ('resource', 'data'):
        for resource in _step_obj.context.stash:
            if not isinstance(resource, dict) \
                    or 'values' not in resource \
                    or 'address' not in resource \
                    or 'type' not in resource:
                resource = {'values': resource,
                            'address': resource,
                            'type': _step_obj.context.name}

            values = resource.get('values', resource.get('expressions', {}))
            if not values:
                values = seek_key_in_dict(resource, something)

            found_values = []
            found_key = Null
            resource_passed = False
            # set this to True if you get anything from the resource, don't set it to False if you get empty values as there could be other values as well
            if isinstance(values, dict):

                found_key = match.get(values, something, Null)
                if found_key is not Null:
                    found_key = [{something: found_key}]
                else:
                    found_key = seek_key_in_dict(values, something)

                for kv_pair in found_key:
                    # kv_pair must be in {something: found_key} format.
                    if not isinstance(kv_pair, dict):
                        continue # could raise an exception
                    # ignore the values that correspond to Null 
                    # Following line could be problematic, how to determine if something is set to be empty or not set? Behavior is provider dependent. 
                    # For now, allow '' and don't allow [] as per user cases.
                    if match.get(kv_pair, something) not in ([],):
                        found_values.append(match.get(kv_pair, something))
                        resource_passed = True

            elif isinstance(values, list):

                for value in values:

                    if isinstance(value, dict):
                        # First search in the keys
                        found_key = seek_key_in_dict(value, something)

                        # Then search in the values with 'key'
                        if not found_key:
                            found_key = seek_regex_key_in_dict_values(value, 'key', something)

                            if found_key:
                                found_key = found_key[0]
                                found_values.extend(value.get('value'))
                                resource_passed = True
                                continue
                    elif isinstance(value, list):
                        _, temp_found_values = it_must_contain_something(_step_obj, something, value, child=True)
                        prop_list.extend(temp_found_values)
                        resource_passed = True

                    elif isinstance(value, (str, bool, int, float)):
                        if match.equals(value, something):
                            found_values.append(value)
                            resource_passed = True

                    if found_key is not Null and len(found_key):

                        for found_key_instance in found_key:
                            if isinstance(found_key_instance, dict):
                                if match.get(found_key_instance, something, Null) not in (Null, [], '', {}):
                                    found_values.append(match.get(found_key_instance, something))
                                    resource_passed = True

            for i, found_val in enumerate(found_values):
                if isinstance(found_val, dict) and 'constant_value' in found_val:
                    found_values[i] = found_val['constant_value']

            if resource_passed:
                Error(_step_obj, '{} property exists in {} ({}).'.format(something, resource['address'], resource.get('type', '')))

    elif _step_obj.context.type == 'provider':
        for provider_data in _step_obj.context.stash:
            values = seek_key_in_dict(provider_data, something)

            if values:
                Error(_step_obj, '{} {} does not have {} property.'.format(_step_obj.context.addresses,
                                                                           _step_obj.context.type,
                                                                           something))



from __future__ import unicode_literals

import colorful
import sys
from radish.exceptions import (
    RadishError,
    HookError,
)
from radish.utils import Failure, console_write
from terraform_compliance.common.defaults import Defaults

from radish.exceptions import (
    LanguageNotSupportedError,
    FeatureFileNotFoundError,
    FeatureFileSyntaxError,
    StepDefinitionNotFoundError,
    StepPatternError,
    StepRegexError,
    SameStepError,
    ScenarioNotFoundError,
)


def write_error(text):
    """
        Writes the given text to the console
    """
    console_write("{} {}: {}".format(Defaults().warning_icon,
                                     Defaults().failure_colour("ERROR"),
                                     Defaults().warning_colour(text)))


def write_failure(failure):
    """
        Writes the failure to the console
    """
    console_write("\n{0}".format(Defaults().failure_colour(failure.traceback)))


def abort(return_code):
    """
        Aborts the program with the given return_code
    """
    sys.exit(return_code)


def handle_exception(exception):
    """
        Handle the given exception

        This will print more information about the given exception

        :param Exception exception: the exception to handle
    """
    if isinstance(exception, HookError):
        write_error(exception)
        write_failure(exception.failure)
        abort(1)
    elif isinstance(exception, RadishError):
        write_error(handle_radish_errors(exception))
        abort(1)
    elif isinstance(exception, KeyboardInterrupt):
        console_write("Aborted by the user...")
        abort(1)
    else:
        write_error(exception)
        write_failure(Failure(exception))
        abort(2)


def handle_radish_errors(error_text):
    mapping = [
        {
            'class': FeatureFileSyntaxError,
            'text': 'You have a SyntaxError in your feature file! Please have a look into terraform-compliance and '
                    'radish documentation to find out which features are supported and how you could use them: '
                    'https://terraform-compliance.com  https://radish.readthedocs.io/en/stable/'
        },
        {
            'class': SameStepError,
            'text': '\n'.join(str(error_text).split('\n')[3:])
        },
        {
            'class': StepDefinitionNotFoundError,
            'text': '{}. Looks like a syntax error.'.format(str(error_text).split('\n'))
        },
    ]

    for map in mapping:
        if map['class'] is error_text.__class__:
            return map['text']

    return error_text



import json
from terraform_compliance.common.helper import seek_key_in_dict, flatten_list, Match, merge_dicts, remove_constant_values
import sys
from copy import deepcopy
from radish.utils import console_write
from terraform_compliance.common.defaults import Defaults
from terraform_compliance.extensions.cache import Cache
from terraform_compliance.common.helper import recursive_jsonify, strip_iterations, get_most_child_module


class TerraformParser(object):
    def __init__(self, filename, parse_it=True):
        '''
        This class reads the given terraform plan filename ( in json format ) and assigns required variables for
        further steps in terraform-compliance. If the file is not a json or terraform plan file, then it will be
        checked and exited in prior steps.

        :param filename: terraform plan filename in json format.
        :parse_it: Runs self.parse() if given.

        :return: None
        '''
        self.supported_terraform_versions = (
            '0.12.',
            '0.13.',
            '0.14.',
            '0.15.',
            '1.0.',
            '1.1.',
            '1.2.',
            '1.3.',
        )
        self.supported_format_versions = [
            '0.1',
            '0.2',
            '1.0',
            '1.1',
            '1.2',
        ]

        self.raw = self._read_file(filename)
        self.variables = None
        self.resources = {}

        self.data = {}

        self.providers = {}

        self.configuration = dict(resources={}, variables={})
        self.file_type = "plan"
        self.resources_raw = {}
        self.type_to_after_unknown_properties = {}
        self.parse_it = parse_it

        if parse_it:
            self.cache = Cache()
            self.parse()

    def _version_check(self):
        if self.raw['format_version'] not in self.supported_format_versions:
            print('\nFATAL ERROR: Unsupported terraform plan output format version '
                  '({}).\n'.format(self.raw['format_version']))
            sys.exit(1)

        if not self.raw['terraform_version'].startswith(self.supported_terraform_versions):
            print('\nFATAL ERROR: Unsupported terraform version '
                  '({}).\n'.format(self.raw['terraform_version']))
            sys.exit(1)

        return True

    def _identify_data_file(self):
        if 'values' in self.raw:
            self.file_type = 'state'

    def _read_file(self, filename):
        '''
        Reads the json filename as a dictionary. We are not checking if the file is a json file again, since
        it is already checked in main.py

        :param filename: json filename with full path
        :return: parsed dictionary
        '''
        with open(filename, 'r', encoding='utf-8') as plan_file:
            data = json.load(plan_file)

        return data

    def _parse_variables(self):
        '''
        Assignes all variables that is defined within the terraform plan

        :return: none
        '''
        self.variables = self.raw.get('variables', {})

    def _parse_resources(self):
        '''
        Assigns all resources defined in the terraform plan

        :return: none
        '''

        # Read Cache
        if self.parse_it:
            cache = self.cache.get('resources')
            if cache:
                self.resources = cache
                return

        # Resources ( exists in Plan )
        for findings in seek_key_in_dict(self.raw.get('planned_values', {}).get('root_module', {}), 'resources'):
            for resource in findings.get('resources', []):
                if self.is_type(resource, 'data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        # Resources ( exists in State )
        for findings in seek_key_in_dict(self.raw.get('values', {}).get('root_module', {}), 'resources'):
            for resource in findings.get('resources', []):
                if self.is_type(resource, 'data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        # Resources ( exists in Prior State )
        for findings in seek_key_in_dict(self.raw.get('prior_state', {}).get('values', {}).get('root_module', {}).get('resources', {}), 'resource'):
            for resource in findings.get('resources', []):
                if self.is_type(resource, 'data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        # Child Modules Resources ( exists in State )
        for findings in seek_key_in_dict(self.raw.get('values', {}).get('root_module', {}), 'child_modules'):
            for resource in findings.get('resources', []):
                if self.is_type(resource, 'data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        # Resource Changes ( exists in Plan )
        for finding in self.raw.get('resource_changes', {}):
            resource = deepcopy(finding)
            change = resource.get('change', {})
            actions = change.get('actions', [])
            if actions != ['delete']:
                resource['values'] = change.get('after', {}) # dict_merge(change.get('after', {}), change.get('after_unknown', {}))
                self.remember_after_unknown(resource, change.get('after_unknown', {}))
                if 'change' in resource:
                    del resource['change']

                if self.is_type(resource, 'data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        if self.parse_it:
            self.cache.set('resources', self.resources)

    def remember_after_unknown(self, resource, after_unknown):
        '''
        Creates a map of resource type to after_unknown values
        These can be used in given step "resource that supports x"

        Note: This function may be extended to capture 'after' values as well. That would require flattening the multi level
        dictionaries in resource
        '''

        # get type
        resource_type = resource.get('type', '')

        # if resource doesn't have type field, try to extract it from address (ideally, this if never evaluates true)
        if not resource_type and 'address' in resource and resource['address']:
            parsed_address = resource.get('address').split('.')
            if parsed_address != 'module':
                resource_type = parsed_address[0]
            elif len(parsed_address) >= 3:
                resource_type = parsed_address[2]
            else:
                return

        # get after_unknown values

        # need to merge because which values are in after_unknown may change from instance to instance
        # merging rule: if field not in map, add it
        if resource_type not in self.type_to_after_unknown_properties:
            self.type_to_after_unknown_properties[resource_type] = after_unknown
        else:
            for key, value in after_unknown.items():
                if key not in self.type_to_after_unknown_properties[resource_type]:
                    self.type_to_after_unknown_properties[resource_type][key] = value

    def _parse_configurations(self):
        '''
        Assigns all configuration related data defined in the terraform plan. This is mostly used for
        resources referencing each other.

        :return: none
        '''

        # Read Cache
        if self.parse_it:
            cache = self.cache.get('configuration')
            if cache:
                self.configuration = cache
                return

        # Resources
        self.configuration['resources'] = {}

        # root resources
        resources = self.raw.get('configuration', {}).get('root_module', {}).get('resources', [])

        # Append module resources
        resources.extend(self.process_module_calls(self.raw.get('configuration', {}).get('root_module', {}).get("module_calls", {})))

        remove_constant_values(resources)
        for resource in resources:
            if self.is_type(resource, 'data'):
                self.data[resource['address']] = resource

            else:
                self.configuration['resources'][resource['address']] = resource

        # Variables
        self.configuration['variables'] = {}
        for findings in seek_key_in_dict(self.raw.get('configuration', {}).get('root_module', {}), 'variables'):
            self.configuration['variables'] = findings.get('variables')

        # Providers
        self.configuration['providers'] = {}
        for findings in seek_key_in_dict(self.raw.get('configuration', {}), 'provider_config'):
            self.configuration['providers'] = findings.get('provider_config', {})

        # Outputs
        self.configuration['outputs'] = {}
        for findings in seek_key_in_dict(self.raw.get('configuration', {}), 'outputs'):
            for key, value in findings.get('outputs', {}).items():
                tmp_output = dict(address=key, value={})
                if 'expression' in value:
                    if 'references' in value['expression']:
                        tmp_output['value'] = value['expression']['references']
                        tmp_output['type'] = 'object'
                    elif 'constant_value' in value['expression']:
                        tmp_output['value'] = value['expression']['constant_value']

                if 'sensitive' in value:
                    tmp_output['sensitive'] = str(value['sensitive']).lower()
                else:
                    tmp_output['sensitive'] = 'false'

                if 'type' in value:
                    tmp_output['type'] = value['type']
                elif 'type' not in tmp_output:
                    if isinstance(tmp_output['value'], list):
                        tmp_output['type'] = 'list'
                    elif isinstance(tmp_output['value'], dict):
                        tmp_output['type'] = 'map'
                    elif isinstance(tmp_output['value'], str):
                        tmp_output['type'] = 'string'
                    elif isinstance(tmp_output['value'], int):
                        tmp_output['type'] = 'integer'
                    elif isinstance(tmp_output['value'], bool):
                        tmp_output['type'] = 'boolean'

                self.configuration['outputs'][key] = tmp_output

        if self.parse_it:
            self.cache.set('configuration', self.configuration)

    def _mount_resources(self, source, target, ref_type):
        '''
        Mounts values of the source resource to the target resource's values with ref_type key

        :param source: source resource
        :param target: target resource
        :param ref_type: reference type (e.g. ingress )
        :return: none
        '''
        for source_resource in source:

            if 'values' not in self.resources.get(source_resource, {}):
                continue
            for parameter, target_resources in target.items():
                for target_resource in target_resources:
                    if target_resource not in self.resources or 'values' not in self.resources[target_resource]:
                        continue

                    resource = self.resources_raw[source_resource]['values']

                    # This is a very stupid terraform-provider bug. Somehow, sometimes it loses the state
                    # and sets the value to None - which is normally not allowed.. It should have been an empty
                    # dict instead. Hence, we are fixing that here.
                    if resource is None:
                        defaults = Defaults()
                        console_write('{} {}: {}'.format(defaults.warning_icon,
                                                         defaults.warning_colour('WARNING (mounting)'),
                                                         defaults.info_colour('The resource "{}" has no values set. This is a terraform provider '
                                                                              'bug. Its recommended to remove/fix this resource within your state.'.format(source_resource))))
                        self.resources_raw[source_resource]['values'] = {}
                        self.resources[source_resource]['values'] = {}
                        resource = {}

                    resource[Defaults.mounted_ptr] = True

                    if Defaults.r_mount_ptr not in self.resources[target_resource]:
                        self.resources[target_resource][Defaults.r_mount_ptr] = {}

                    if Defaults.r_mount_addr_ptr not in self.resources[target_resource]:
                        self.resources[target_resource][Defaults.r_mount_addr_ptr] = {}

                    if Defaults.r_mount_addr_ptr_list not in self.resources[target_resource]:
                        self.resources[target_resource][Defaults.r_mount_addr_ptr_list] = []

                    # ensure resources[target_resource]['values'] is an
                    # empty dict and not None
                    if not self.resources[target_resource]['values']:
                        self.resources[target_resource]['values'] = dict()

                    if ref_type not in self.resources[target_resource]['values']:
                        self.resources[target_resource]['values'][ref_type] = []

                    self.resources[target_resource]['values'][ref_type].append(resource)
                    self.resources[target_resource][Defaults.r_mount_ptr][parameter] = ref_type
                    self.resources[target_resource][Defaults.r_mount_addr_ptr][parameter] = source
                    target_set = set(self.resources[target_resource][Defaults.r_mount_addr_ptr_list])
                    source_set = set(source)
                    self.resources[target_resource][Defaults.r_mount_addr_ptr_list] = list(target_set | source_set)

                    if parameter not in self.resources[source_resource]['values']:
                        self.resources[source_resource]['values'][parameter] = target_resource

    def _find_resource_from_name(self, resource_name, module_address=None):
        '''
        Finds all the resources that is starting with resource_name

        :param resource_name: The first initials of the resource
        :param module_address: Full module address (without the resource)
        :return: list of the found resources
        '''
        if resource_name in self.resources:
            return [resource_name]

        resource_list = []
        # Try to find the resource with the module address in self.resources
        if module_address is not None:
            full_address = '{}.{}'.format(module_address, resource_name)
            if full_address in self.resources:
                return [full_address]

            for key, value in self.resources.items():
                if not key.startswith(module_address):
                    continue

                # Check if the resource/module has iterations
                if '[' in key:
                    # Possibly module (or resource) is using foreach/count
                    k = strip_iterations(key)
                    if k == strip_iterations(full_address):
                        resource_list.append(key)
                else:
                    # Resource/module does not have any iteration
                    # Additionally, the resource we are looking under this module is coming from another
                    # module output. Thus, we need to dive a bit deeper.
                    if resource_name.startswith('module') and key.startswith('module'):
                        resource_list.append(key)

            if resource_list:
                return resource_list

        resource_type, resource_id = resource_name.split('.')[0:2]

        if resource_type == 'module':
            # TODO: This wont work correctly, if an output is used within a module, coming from another module.
            #       Fix multi-layer module structure for the outputs ?
            module_name, output_id = resource_name.split('.')[1:3]
            module = self.raw['configuration']['root_module'].get('module_calls', {}).get(module_name, {})

            output_value = module.get('module', {}).get('outputs', {}).get(output_id, {})

            resources = output_value.get('expression', {}).get('references', []) if 'expression' in output_value else output_value.get('value', [])

            resources = ['{}.{}.{}'.format(resource_type, module_name, res) for res in resources]

            if not resources:
                for key, _ in self.resources.items():
                    if key.startswith(resource_name):
                        resources.append(key)

            if resources:
                resource_list.extend(resources)
        else:
            for key, value in self.resources.items():
                if value['type'] == resource_type and value['name'] == resource_id:
                    resource_list.append(key)

        return resource_list

    def _mount_references(self):
        '''
        Find the references that is defined in self.configuration
        :return:
        '''
        self.resources_raw = deepcopy(self.resources)
        invalid_references = ('var.', 'each.', 'count.')

        # This section will link resources found in configuration part of the plan output.
        # The reference should be on both ways (A->B, B->A) since terraform sometimes report these references
        # in opposite ways, depending on the provider structure.
        for resource in self.configuration['resources']:
            relative_resource_address = '{}.{}'.format(self.configuration['resources'][resource]['type'], self.configuration['resources'][resource]['name'])
            current_module_address = self.configuration['resources'][resource]['address'].replace('.{}'.format(relative_resource_address), '')

            if 'expressions' in self.configuration['resources'][resource]:
                ref_list = {}

                for key, value in self.configuration['resources'][resource]['expressions'].items():
                    references = seek_key_in_dict(value, 'references') if isinstance(value, (dict, list)) else []

                    valid_references = []
                    for ref in references:
                        if isinstance(ref, dict) and ref.get('references'):
                            valid_references = []
                            for r in ref['references']:
                                if r.startswith('var'):
                                    # Try to track the resource given by a variable
                                    _var = self._fetch_resource_by_a_variable(current_module_address, r)
                                    if _var:
                                        valid_references.extend(_var)

                                if not r.startswith(invalid_references):
                                    valid_references.append(r)

                    for ref in valid_references:
                        # if ref is not in the correct format, handle it
                        if len(ref.split('.')) < 3 and ref.startswith('module'):

                            # Using for_each and modules together may introduce an issue where the plan.out.json won't
                            # include the necessary third part of the reference. It is partially resolved by mounting
                            # the reference to all instances belonging to the module
                            if 'for_each_expression' in self.configuration['resources'][resource]:

                                # extract source resources
                                assumed_source_resources = [k for k in self.resources.keys() if k.startswith(resource)]
                                # extract for_each keys
                                assumed_for_each_keys = [k[len(resource):].split('.')[0] for k in assumed_source_resources]
                                # combine ref with for each keys
                                assumed_refs = ['{}{}'.format(ref, key) for key in assumed_for_each_keys]
                                # get all the resources that start with updated ref
                                ambiguous_references = []
                                for r in self.resources.keys():
                                    for assumed_ref in assumed_refs:
                                        if r.startswith(assumed_ref):
                                            if key in ref_list:
                                                ref_list[key].append(r)
                                            else:
                                                ref_list[key] = [r]

                                            ambiguous_references.append(r)

                                # throw a warning
                                defaults = Defaults()
                                console_write('{} {}: {}'.format(defaults.warning_icon,
                                       defaults.warning_colour('WARNING (mounting)'),
                                       defaults.info_colour('The reference "{}" in resource {} is ambigious.'
                                        ' It will be mounted to the following resources:').format(ref, resource)))
                                for i, r in enumerate(ambiguous_references, 1):
                                    console_write(defaults.info_colour('{}. {}'.format(i, r)))

                            # if the reference can not be resolved, warn the user and continue.
                            else:
                                console_write('{} {}: {}'.format(Defaults().warning_icon,
                                       Defaults().warning_colour('WARNING (mounting)'),
                                       Defaults().info_colour('The reference "{}" in resource {} is ambigious. It will not be mounted.'.format(ref, resource))))
                                continue
                        elif key not in ref_list:
                            ref_list[key] = self._find_resource_from_name(ref, current_module_address)
                        else:
                            ref_list[key].extend(self._find_resource_from_name(ref, current_module_address))

                    # This is where we synchronise constant_value in the configuration section with the resource
                    # for filling up the missing elements that hasn't been defined in the resource due to provider
                    # implementation.
                    target_resource = [t for t in [self.resources.get(resource, {}).get('address')] if t is not None]
                    if not target_resource:
                        target_resource = [k for k in self.resources.keys() if k.startswith(resource)]

                    for t_r in target_resource:
                        if self.resources[t_r].get('values') is None:
                            continue
                        if type(value) is type(self.resources[t_r]['values'].get(key)) and self.resources[t_r]['values'].get(key) != value:
                            if isinstance(value, (list, dict)):
                                merge_dicts(self.resources[t_r]['values'][key], value)

                if ref_list:
                    ref_type = self.configuration['resources'][resource]['expressions'].get('type', {})

                    if 'references' in ref_type:
                        ref_type = resource.split('.')[0]

                    if not ref_type and not self.is_type(resource, 'data'):
                        ref_type = self.extract_resource_type_from_address(resource)

                    for k, v in ref_list.items():
                        v = flatten_list(v)

                    # Mounting A->B
                    source_resources = self._find_resource_from_name(self.configuration['resources'][resource]['address'])

                    # Try again in case we might have for_each/count usage for the module
                    if not source_resources:
                        source_resources = self._find_resource_from_name(relative_resource_address, current_module_address)

                    self._mount_resources(source=source_resources,
                                          target=ref_list,
                                          ref_type=ref_type)

                    # Mounting B->A
                    for parameter, target_resources in ref_list.items():
                        for target_resource in target_resources:
                            if not self.is_type(resource, 'data') and not self.is_type(resource, 'var') and not self.is_type(resource, 'provider'):
                                ref_type = self.extract_resource_type_from_address(target_resource)

                                self._mount_resources(source=[target_resource],
                                                      target={parameter: source_resources},
                                                      ref_type=ref_type)

    def _distribute_providers(self):
        for resource_name, resource_data in self.resources.items():
            resource_provider = resource_name.split('_')[0]

            if resource_provider not in self.providers:
                self.providers[resource_provider] = {}

            self.providers[resource_provider][resource_name] = resource_data

    def parse(self):
        '''
        Main method for initialising the parsing of the terraform plan json file

        :return: nothing
        '''
        self._version_check()
        self._identify_data_file()
        self._parse_resources()

        if self.file_type == 'plan':
            self._parse_variables()
            self._parse_configurations()

        cache_mounted_resources = self.cache.get('mounted_resources') if self.parse_it else None
        cache_raw_resources = self.cache.get('resources_raw') if self.parse_it else None
        cache_type_to_after_unknown_properties = self.cache.get('type_to_after_unknown_properties') if self.parse_it else None

        if cache_mounted_resources and cache_raw_resources:
            # print('Read from cache, instead of re-mounting.')
            self.resources = cache_mounted_resources
            self.resources_raw = cache_raw_resources
            self.type_to_after_unknown_properties = cache_type_to_after_unknown_properties
        else:
            # print('Building cache for mounted resources at {}'.format(Defaults.cache_dir))
            self._mount_references()
            # metadata related calls
            self._add_action_status()
            self._add_module_call_source()

            self.resources = recursive_jsonify(self.resources)
            self.resources_raw = recursive_jsonify(self.resources_raw)
            self.type_to_after_unknown_properties = recursive_jsonify(self.type_to_after_unknown_properties)
            self.variables = recursive_jsonify(self.variables)
            self.data = recursive_jsonify(self.data)
            self.providers = recursive_jsonify(self.providers)

            if self.parse_it:
                self.cache.set('mounted_resources', self.resources)
                self.cache.set('resources_raw', self.resources_raw)
                self.cache.set('type_to_after_unknown_properties', self.type_to_after_unknown_properties)

        self._distribute_providers()

        for _, resource in self.resources.items():
            self._expand_resource_tags(resource)

    def _add_action_status(self):
        '''
        Adds Terraform's action status to each resource
        '''
        if 'resource_changes' not in self.raw:
            return

        for resource_change in self.raw['resource_changes']:
            resource = resource_change['address']
            if resource in self.resources:
                self.resources[resource]['actions'] = resource_change['change']['actions']

    def _add_module_call_source(self):
        '''
        Adds module call's source to module's resources as metadata
        '''
        for resource in self.resources.values():
            # removes the for_each signature from addresses
            # module.a["index_1"].b.c -> module.a.b.c
            fixed_module_name = '.'.join([word.split('[')[0] for word in resource['address'].split('.')])

            if 'source' in self.configuration['resources'].get(fixed_module_name, ''):
                resource['source'] = self.configuration['resources'][fixed_module_name]['source']


    def find_resources_by_type(self, resource_type, match=Match(case_sensitive=False)):
        '''
        Finds all resources matching with the resource_type

        :param resource_type: String of resource type defined in terraform
        :return: list of dict including resources
        '''
        resource_list = []

        for resource_data in self.resources.values():
            if resource_type == 'any' or (match.equals(resource_data['type'], resource_type) and resource_data['mode'] == 'managed'):
                resource_list.append(resource_data)

        return resource_list

    def find_data_by_type(self, resource_type, match=Match(case_sensitive=False)):
        '''
        Finds all data matching with the resource_type

        :param resource_type: String of resource type defined in terraform
        :return: list of dict including resources
        '''
        resource_list = []

        for resource_data in self.data.values():
            if match.equals(resource_data['type'], resource_type):
                resource_list.append(resource_data)

        return resource_list

    def get_providers_from_configuration(self, provider_type, match=Match(case_sensitive=False)):
        '''
        Returns all providers as a list for the given provider type

        :param provider_type: String of a provider type like aws
        :return: list of providers that has this type
        '''
        providers = []
        for provider_alias, values in self.configuration['providers'].items():
            if isinstance(values, dict) and match.equals(values.get('name'), provider_type):
                providers.append(values)

        return providers

    def _expand_resource_tags(self, resource):
        if isinstance((resource.get('values') or {}).get('tags'), list):
            for tag in resource.get('values', {}).get('tags', {}):
                if isinstance(tag, dict) and 'key' in tag and 'value' in tag:
                    tag[tag['key']] = tag['value']
            return True
        return False

    def is_type(self, resource, mode):
        if isinstance(resource, dict):
            if 'mode' in resource:
                return resource['mode'] == mode

            return resource['address'].split('.')[0] == mode

        return False

    def process_module_calls(self, module_resource, parents_modules=None):
        '''
        This method will recursively process modules and extract resources from "module_resource" data
        which is actually a data from self.configuration dict. We were returning the native resource name
        before this method, but now we are returning proper address naming for the resource.

        :param module_resource: The self.configuration part
        :param parents_modules: internal usage for recursive functionality
        :return: None
        '''
        if parents_modules is None:
            parents_modules = []

        resources = []
        for k, v in module_resource.items():
            # Set the naming correct (for cases like module.a.module.b.module.c...)
            current_module_level = deepcopy(parents_modules)
            current_module_level.append('module.{}'.format(k))
            module_name = ".".join(current_module_level)
            # Pull module's source to be later used in metadata
            module_source = v.get('source', '')

            # Register the resource (along with module naming)
            if 'resources' in v.get('module', {}):
                for resource in v['module']['resources']:
                    resource['address'] = '{}.{}'.format(module_name, resource['address'])
                    resource['source'] = module_source
                    resources.append(resource)

            # Dive deeper, its not finished yet.
            if 'module_calls' in v.get('module', {}):
                resources.extend(self.process_module_calls(v['module']['module_calls'], current_module_level))

        return resources

    def extract_resource_type_from_address(self, resource_address_string):
        '''
        Tries to get the resource type from the resource address

        :param resource_address_string: String of the whole resource address
        :return: String of the resource type if found, otherwise will return full address

        Example;

            "aws_s3_bucket.test" will return "aws_s3_bucket"
            "module_a.module_b.module_c.aws_s3_bucket.test" will return "aws_s3_bucket"
            "something_else" will return "something_else"
        '''
        if '.' in resource_address_string:
            octets = resource_address_string.split('.')
            if len(octets) > 1:
                # Return the type as we found it properly
                return octets[-2]
            else:
                # Return the whole address
                return octets[0]

        # Returning the whole address
        return resource_address_string

    def _fetch_resource_by_a_variable(self, module, variable):
        target_module = get_most_child_module(module)
        stripped_variable = variable.replace('var.', '')
        var = self.raw['configuration'].get('root_module', {}).get('module_calls', {}).get(target_module, {}).get('expressions', {}).get(stripped_variable, {}).get('references', {})
        return var



# -*- coding: utf-8 -*-

from terraform_compliance.common.helper import (
    EmptyStash,
    get_resource_name_from_stash
)
from terraform_compliance.common.error_handling import Error
import re


def its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, _stash=EmptyStash):
    def fail(condition, name=None):
        text = 'matches' if condition == 'must not' else 'does not match'
        name = name if (name is not None or name is not False) else _step_obj.context.name
        pattern = 'Null/None' if regex == '\x00' else regex
        Error(_step_obj, '{} property in {} {} {} with {} {} regex. '
                         'It is set to {}.'.format(_step_obj.context.property_name,
                                                   name,
                                                   _step_obj.context.type,
                                                   text,
                                                   pattern,
                                                   regex_flag_error_text,
                                                   values))

    match = _step_obj.context.match
    regex = r'{}'.format(search_regex)
    values = _step_obj.context.stash if _stash is EmptyStash else _stash
    regex_flag_error_text = 'case insensitive' if not match.case_sensitive else 'case sensitive'

    if isinstance(values, (str, int, bool, float)) or values is None:
        matches = match.regex_match(regex, str(values))

        if (condition == 'must' and matches is None) or (condition == "must not" and matches is not None):
            _stash = get_resource_name_from_stash(_step_obj.context.stash, _stash, _step_obj.context.address)
            fail(condition, name=_stash.get('address'))

    elif isinstance(values, list):
        for value in values:
            its_value_condition_match_the_search_regex_regex(_step_obj,
                                                             condition,
                                                             search_regex,
                                                             value)

    elif isinstance(values, dict):
        if not hasattr(_step_obj.context, 'address'):
            _step_obj.context.address = None

        _step_obj.context.address = values.get('address', _step_obj.context.address)

        if 'values' in values:
            if values['values'] is None and regex == '\x00' and condition == 'must not':
                values = values['values']
                fail(condition, name=_stash.get('address'))
            else:
                its_value_condition_match_the_search_regex_regex(_step_obj,
                                                                 condition,
                                                                 search_regex,
                                                                 values.get('values'))

        else:
            for key, value in values.items():
                its_value_condition_match_the_search_regex_regex(_step_obj,
                                                                 condition,
                                                                 search_regex,
                                                                 value)


def any_of_its_values_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, _stash=EmptyStash):
    def fail(condition, name=None):
        text = 'matches' if condition == 'must not' else 'does not match'
        name = name if (
            name is not None or name is not False) else _step_obj.context.name
        pattern = 'Null/None' if regex == '\x00' else regex
        Error(_step_obj, '{} property in {} {} {} with {} {} regex. '
                         'It is set to {}.'.format(_step_obj.context.property_name,
                                                   name,
                                                   _step_obj.context.type,
                                                   text,
                                                   pattern,
                                                   regex_flag_error_text,
                                                   values))

    found = False
    def search(values):
        nonlocal found
        if found:
            return True

        if isinstance(values, (str, int, bool, float)) or values is None:
            matches = match.regex_match(regex, str(values))

            if (condition == 'must' and matches is not None) or (condition == "must not" and matches is None):
                found = True
                return found

        elif isinstance(values, list):
            return any(map(search, values))

        elif isinstance(values, dict):
            if not hasattr(_step_obj.context, 'address'):
                _step_obj.context.address = None

            _step_obj.context.address = values.get('address', _step_obj.context.address)

            if 'values' in values:
                return search(values['values'])
            else:
                return any(map(search, values.values()))

        return False

    match = _step_obj.context.match
    regex = r'{}'.format(search_regex)
    values = _step_obj.context.stash if _stash is EmptyStash else _stash
    regex_flag_error_text = 'case insensitive' if not match.case_sensitive else 'case sensitive'

    if not search(values):
        _stash = get_resource_name_from_stash(_step_obj.context.stash, _stash, _step_obj.context.address)
        fail(condition, name=_stash.get('address'))


def its_singular_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, _stash=EmptyStash):
    resources = _step_obj.context.stash if _stash is EmptyStash else _stash

    if isinstance(resources, dict): 
        if 'values' in resources: # in case the object is in 'address', 'values', 'type' format
            resources = resources['values']
        else:
            Error(_step_obj, '{} is multivalued! Please use any/all versions of this step instead.'.format(_step_obj.context.property_name,))
            return

    elif isinstance(resources, list):
        for resource in resources:
            if isinstance(resource, dict):
                resource = resource.get('values', resource)

            if isinstance(resource, (dict, list)) and len(resource) > 1:
                Error(_step_obj, '{} is multivalued! Please use any/all versions of this step instead.'.format(_step_obj.context.property_name,))
                return

    its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, _stash)



from unittest import TestCase
from terraform_compliance.common.terraform_files import (
    which,
    convert_terraform_plan_to_json,
    download_terraform,
    get_platform_details,
    detect_required_terraform_version
)

import os
from mock import patch


class TestTerraformFiles(TestCase):

    def test_which_success(self):
        self.assertEqual('/bin/ls', which('/bin/ls'))

    def test_which_failure(self):
        self.assertIsNone(which('/something/that/doesnt/exist/hopefully'))

    @patch('terraform_compliance.common.terraform_files.which', return_value=None)
    def test_convert_terraform_plan_to_json(self, *args):
        with self.assertRaises(SystemExit):
            convert_terraform_plan_to_json('some_file')

    def test_get_platform_details(self):
        system, arch = get_platform_details()
        print("System: {}\tArch: {}".format(system, arch))
        self.assertIsNot(system, "")
        self.assertIsNot(arch, "")

    # def test_downloading_terraform(self, *args):
    #     tf = download_terraform('0.15.3')

    def test_detect_required_terraform_version(self):
        current, target = detect_required_terraform_version('0.14.0, but this is 0.15.1; plan files cannot be transferred between')
        self.assertEqual(current, '0.14.0')
        self.assertEqual(target, '0.15.1')

        current, target = detect_required_terraform_version('')
        self.assertIsNone(current)
        self.assertIsNone(target)


from unittest import TestCase
from tests.mocks import MockedData, MockedWorld
from terraform_compliance.extensions.security_groups import SecurityGroup, SecurityGroupRule
from terraform_compliance.common.exceptions import Failure, TerraformComplianceInvalidData
from mock import patch


class TestSecurityGroupRule(TestCase):

    def setUp(self):
        self.rule_fields = ['cidr_blocks', 'from_port', 'to_port', 'protocol']

    def validate_rule(self, mocked_rule, rule):
        for field in self.rule_fields:
            self.assertEqual(mocked_rule[field],
                             rule.__getattribute__(field))

    def test_assign_sg_params_one_port_with_two_cidrs(self):
        self.validate_rule(MockedData.sg_params_ssh_with_2_cidrs,
                           SecurityGroupRule(**MockedData.sg_ssh_with_2_cidrs))

    def test_assign_sg_params_one_port_two_cidrs_any_proto(self):
        self.validate_rule(MockedData.sg_params_ssh_with_2_cidrs_any_proto,
                           SecurityGroupRule(**MockedData.sg_ssh_with_2_cidrs_any_proto))

    def test_assign_sg_params_all_ports_with_all_ips(self):
        self.validate_rule(MockedData.sg_params_all_port_all_ip,
                           SecurityGroupRule(**MockedData.sg_all_port_all_ip))

    def test_assign_sg_params_no_data_given_in_rules(self):
        self.validate_rule(MockedData.sg_params_all_port_no_ip,
                           SecurityGroupRule(**MockedData.sg_all_port_no_ip))

    def test_assign_sg_params_from_port_bigger_than_to_port(self):
        with self.assertRaises(TerraformComplianceInvalidData) as context:
            SecurityGroupRule(**MockedData.sg_invalid)

            self.assertTrue('Invalid configuration from_port can not be bigger than to_port.' in context.exception)

    # TODO: Implement failure tests here for consistency.


class TestSecurityGroup(TestCase):

    def setUp(self):
        self.sg_given = dict(
            port=80,
            protocol='tcp',
            cidr_blocks='0.0.0.0/0',
        )
        self.sg_in_conf = [
            dict(
                from_port=80,
                to_port=80,
                protocol='tcp',
                cidr_blocks=['0.0.0.0/0'],
                description='Test Security Group Description #1',
                values=dict(address='test.security_group_rule1')
            ),
            dict(
                from_port=81,
                to_port=81,
                protocol='tcp',
                cidr_blocks=['10.0.0.0/8', '192.168.0.0/24'],
                description='Test Security Group Description #2',
                values=dict(address='test.security_group_rule2')

            )
        ]

    # Tests about `must not have` scenarios
    # Checks are singular per security group rule, failures are imminent.
    def test_must_not_have_port_tcp_80_with_ALL_cidr(self):
        result, error = SecurityGroup(self.sg_given, self.sg_in_conf).validate()

        self.assertFalse(result)
        self.assertEqual('tcp/80 port is defined within 0.0.0.0/0 network in test.security_group_rule1.', error)

    def test_must_not_have_port_tcp_80_with_multi_cidr(self):
        self.sg_given['cidr_blocks'] = '192.168.1.0/24'
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16', '0.0.0.0/0']

        result, error = SecurityGroup(self.sg_given, self.sg_in_conf).validate()

        self.assertFalse(result)
        self.assertEqual('tcp/80 port is defined within 192.168.0.0/16, 0.0.0.0/0 networks in '
                         'test.security_group_rule1.', error)

    def test_must_not_have_port_tcp_80_with_ALL_cidr_success(self):
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16']
        self.assertTrue(SecurityGroup(self.sg_given, self.sg_in_conf).validate())

    def test_must_not_have_port_tcp_80_with_ALL_multi_success(self):
        self.sg_given['cidr_blocks'] = '192.168.1.0/16'
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/24', '10.0.0.0/8']
        self.assertTrue(SecurityGroup(self.sg_given, self.sg_in_conf).validate())

    def test_must_not_have_port_tcp_22_23_with_ALL_cidr(self):
        self.sg_given['port'] = '22-23'
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23

        result, error = SecurityGroup(self.sg_given, self.sg_in_conf).validate()

        self.assertFalse(result)
        self.assertEqual('tcp/(22,23) ports are defined within 0.0.0.0/0 network in test.security_group_rule1.', error)

    def test_must_not_have_port_tcp_22_23_with_multi_cidr(self):
        self.sg_given['port'] = '22-23'
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23
        self.sg_given['cidr_blocks'] = '192.168.1.0/24'
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16', '0.0.0.0/0']

        result, error = SecurityGroup(self.sg_given, self.sg_in_conf).validate()

        self.assertFalse(result)
        self.assertEqual('tcp/(22,23) ports are defined within 192.168.0.0/16, 0.0.0.0/0 networks in '
                         'test.security_group_rule1.', error)

    def test_must_not_have_port_tcp_22_23_with_ALL_cidr_success(self):
        self.sg_given['port'] = '22-23'
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16']
        self.assertTrue(SecurityGroup(self.sg_given, self.sg_in_conf).validate())

    def test_must_not_have_port_tcp_22_23_with_ALL_multi_success(self):
        self.sg_given['port'] = '22-23'
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23
        self.sg_given['cidr_blocks'] = '192.168.1.0/16'
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/24', '10.0.0.0/8']
        self.assertTrue(SecurityGroup(self.sg_given, self.sg_in_conf).validate())

    def test_must_not_have_port_tcp_22_with_range_with_ALL_cidr(self):
        self.sg_given['port'] = '21-22'
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23

        result, error = SecurityGroup(self.sg_given, self.sg_in_conf).validate()

        self.assertFalse(result)
        self.assertEqual('tcp/22 port is defined within 0.0.0.0/0 network in test.security_group_rule1.', error)

    def test_must_not_have_port_tcp_22_with_range_with_multi_cidr(self):
        self.sg_given['port'] = '21-22'
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23
        self.sg_given['cidr_blocks'] = '192.168.1.0/24'
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16', '0.0.0.0/0']

        result, error = SecurityGroup(self.sg_given, self.sg_in_conf).validate()

        self.assertFalse(result)
        self.assertEqual('tcp/22 port is defined within 192.168.0.0/16, 0.0.0.0/0 networks in '
                         'test.security_group_rule1.', error)

    def test_must_not_have_port_tcp_22_with_range_with_ALL_cidr_success(self):
        self.sg_given['port'] = '21-22'
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16']
        self.assertTrue(SecurityGroup(self.sg_given, self.sg_in_conf).validate())

    def test_must_not_have_port_tcp_22_with_range_with_ALL_multi_success(self):
        self.sg_given['port'] = '21-22'
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23
        self.sg_given['cidr_blocks'] = '192.168.1.0/16'
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/24', '10.0.0.0/8']
        self.assertTrue(SecurityGroup(self.sg_given, self.sg_in_conf).validate())

    # Tests about `must have` scenarios
    def test_must_have_port_tcp_443_with_ALL_cidr(self):
        self.sg_given['port'] = 443

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_have()
        result, error = sg.validate()

        self.assertFalse(result)
        self.assertEqual('tcp/443 port is not defined within 0.0.0.0/0 network in test_sg.', error)

    def test_must_have_port_tcp_443_with_multi_cidr(self):
        self.sg_given['port'] = 443
        self.sg_given['cidr_blocks'] = '192.168.1.0/24'
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16', '0.0.0.0/0']

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_have()
        result, error = sg.validate()

        self.assertFalse(result)
        self.assertEqual('tcp/443 port is not defined within 192.168.1.0/24 network in test_sg.', error)

    def test_must_have_port_tcp_80_with_multi_cidr_success(self):
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/24']
        self.sg_in_conf[1]['cidr_blocks'] = ['192.168.1.0/24']
        self.sg_in_conf[1]['from_port'] = 79
        self.sg_in_conf[1]['to_port'] = 81
        self.sg_given['cidr_blocks'] = '192.168.0.0/24'

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_have()
        self.assertTrue(sg.validate())

    def test_must_have_port_tcp_80_with_multi_cidr_32_success(self):
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/24']
        self.sg_in_conf[1]['cidr_blocks'] = ['192.168.0.0/16']
        self.sg_in_conf[1]['from_port'] = 79
        self.sg_in_conf[1]['to_port'] = 81
        self.sg_given['cidr_blocks'] = '192.168.0.1/32'
        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_have()
        self.assertTrue(sg.validate())

    @patch('radish.world', return_value=MockedWorld)
    def test_must_have_port_tcp_443_444_with_ALL_cidr(self, *args):
        self.sg_given['port'] = '443-444'

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_have()
        result, error = sg.validate()

        self.assertFalse(result)
        self.assertEqual('tcp/(443,444) ports are not defined within 0.0.0.0/0 network in test_sg.', error)

    def test_must_have_port_tcp_80_81_with_ALL_cidr(self):
        self.sg_given['port'] = '80-82'
        self.sg_in_conf[1]['cidr_blocks'] = ['0.0.0.0/0']

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_have()
        result, error = sg.validate()

        self.assertFalse(result)
        self.assertEqual('tcp/82 port is not defined within 0.0.0.0/0 network in test_sg.', error)

    # Tests about `must only have` scenarios
    # We are just checking if tcp/80 is defined for 0.0.0.0/0
    def test_must_only_have_port_tcp_80_with_ALL_cidr_success(self):
        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_only_have()
        self.assertTrue(sg.validate())

    def test_must_only_have_port_tcp_80_81_with_ALL_cidr_success(self):
        self.sg_given['port'] = '80-81'

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_only_have()
        result, error = sg.validate()

        self.assertFalse(result)
        self.assertEqual('tcp/81 port is not defined within 0.0.0.0/0 network in test_sg.', error)

    def test_must_only_have_port_some_ports_are_over_configured(self):
        self.sg_in_conf[0]['from_port'] = 79
        self.sg_in_conf[0]['to_port'] = 81
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16', '0.0.0.0/0']
        self.sg_in_conf[1]['from_port'] = 80
        self.sg_in_conf[1]['to_port'] = 80
        self.sg_in_conf[1]['cidr_blocks'] = ['0.0.0.0/0']

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_only_have()
        result, error = sg.validate()

        self.assertFalse(result)
        self.assertEqual('tcp/(81,79) ports are defined within 0.0.0.0/0 network in test_sg.', error)

    def test_must_only_have_port_not_match_multiple_errors_given(self):
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16', '0.0.0.0/0']
        self.sg_in_conf[1]['from_port'] = 443
        self.sg_in_conf[1]['to_port'] = 444
        self.sg_in_conf[1]['cidr_blocks'] = ['0.0.0.0/0']

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_only_have()
        result, error = sg.validate()

        self.assertFalse(result)
        self.assertTrue('tcp/80 port is not defined within 0.0.0.0/0 network in test_sg.', error)
        self.assertTrue('tcp/(443,444,22,23) ports are defined within 0.0.0.0/0 network in test_sg.', error)
        self.assertTrue('None of the ports given defined within 0.0.0.0/0 network in test_sg.', error)

    def test_must_only_have_port_match_multiple_ports_not_ranges(self):
        self.sg_in_conf[0]['from_port'] = 22
        self.sg_in_conf[0]['to_port'] = 23
        self.sg_in_conf[0]['cidr_blocks'] = ['192.168.0.0/16', '0.0.0.0/0']
        self.sg_in_conf[1]['from_port'] = 443
        self.sg_in_conf[1]['to_port'] = 444
        self.sg_in_conf[1]['cidr_blocks'] = ['0.0.0.0/0']
        self.sg_given['ports'] = '22,23,443,444'

        sg = SecurityGroup(self.sg_given, self.sg_in_conf)
        sg.must_only_have()
        result, error = sg.validate()

        self.assertFalse(result)
        self.assertTrue('tcp/80 port is not defined within 0.0.0.0/0 network in test_sg.', error)
        self.assertTrue('tcp/(443,444,22,23) ports are defined within 0.0.0.0/0 network in test_sg.', error)
        self.assertTrue('None of the ports given defined within 0.0.0.0/0 network in test_sg.', error)



from unittest import TestCase
from terraform_compliance.common.helper import (
    flatten_list,
    check_if_cidr,
    is_ip_in_cidr,
    seek_key_in_dict,
    find_root_by_key,
    are_networks_same,
    convert_resource_type,
    seek_regex_key_in_dict_values,
    jsonify,
    remove_mounted_resources,
    get_resource_name_from_stash,
    dict_merge,
    is_list_of_dict,
    is_key_exist,
    transform_asg_style_tags,
    python_version_check,
    strip_iterations,
    get_most_child_module
)
from ddt import ddt, data, unpack

@ddt
class TestHelperFunctions(TestCase):

    def test_flatten_single_dimensional_list(self):
        a = ['a', 'b', 'c']
        b = ['a', 'b', 'c']

        self.assertEqual(flatten_list(a), b)

    def test_flatten_multi_dimensional_list(self):
        a = ['a', 'b', ['c']]
        b = ['a', 'b', 'c']

        self.assertEqual(flatten_list(a), b)

    def test_flatten_multi_dimensional_nested_list(self):
        a = ['a', 'b', ['c', ['d', 'e'], 'f'], 'g', 'h', 'i', ['j', 'k']]
        b = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k']

        self.assertEqual(flatten_list(a), b)

    def test_check_if_cidr_success(self):
        self.assertTrue(check_if_cidr('10.0.0.0/8'))
        self.assertTrue(check_if_cidr('10.14.0.0/16'))
        self.assertTrue(check_if_cidr('10.0.0.0/24'))
        self.assertTrue(check_if_cidr('10.0.0.7/32'))

    def test_check_if_cidr_failure(self):
        self.assertFalse(check_if_cidr(123))
        self.assertFalse(check_if_cidr(False))
        self.assertFalse(check_if_cidr([]))
        self.assertFalse(check_if_cidr({}))

    def test_is_ip_in_cidr_success(self):
        self.assertTrue(is_ip_in_cidr('10.0.0.0/8', ['0.0.0.0/0']))
        self.assertTrue(is_ip_in_cidr('10.0.0.0/16', ['10.0.0.0/8']))
        self.assertTrue(is_ip_in_cidr('10.0.200.0/24', ['10.0.0.0/16']))
        self.assertTrue(is_ip_in_cidr('10.0.0.1/32', ['10.0.0.0/24']))

    def test_is_ip_in_cidr_failure(self):
        self.assertFalse(is_ip_in_cidr('200.0.0.0/16', ['10.0.0.0/8']))
        self.assertFalse(is_ip_in_cidr('10.200.0.0/24', ['10.0.0.0/16']))
        self.assertFalse(is_ip_in_cidr('10.0.1.1/32', ['10.0.0.0/24']))

    def test_seek_in_dict_finding_a_key_in_root(self):
        dictionary = dict(search_key=dict(something=[]))
        search_key = 'search_key'
        expected = [{'search_key': {'something': []}}]

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_finding_a_key_in_nested_dict(self):
        dictionary = dict(search_key=dict(something=dict(something_else=None)))
        search_key = 'something'
        expected = [{'something': {'something_else': None}}]

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_finding_multiple_keys_in_nested_dict(self):
        dictionary = dict(search_key=dict(something=dict(something_else=None, something=['something_else'])), something=[])
        search_key = 'something'
        expected = [{'something': {'something_else': None, 'something': ['something_else']}}, {'something': []}]

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_finding_values_in_non_dicts(self):
        dictionary = dict(search_key=dict(something=[]))
        search_key = 'something_else'
        expected = []

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_for_root_and_non_root_values(self):
        dictionary = dict(a=dict(b=dict(something=dict(test=[]))), something=["b"])
        search_key = "something"
        expected = [{'something':{'test': []}}, {'something':['b']}]

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_seek_in_dict_finding_values_in_non_dicts_on_root(self):
        dictionary = 'something_else'
        search_key = 'something_else'
        expected = []

        self.assertEqual(seek_key_in_dict(dictionary, search_key), expected)

    def test_find_root_by_key_single_return(self):
        haystack = dict(some_key=dict(values=dict(tags=[], something_else='something')))
        search_key = 'tags'
        expected = ['some_key']

        self.assertEqual(find_root_by_key(haystack, search_key), expected)

    def test_find_root_by_key_multiple_return(self):
        haystack = dict(some_key=dict(values=dict(tags=[], something_else='something')), other_key=dict(values=dict(tags=[], something_else='something')))
        search_key = 'tags'
        expected = ['other_key', 'some_key']

        self.assertEqual(sorted(find_root_by_key(haystack, search_key)), sorted(expected))

    def test_find_root_by_key_multiple_return_02(self):
        haystack = dict(some_key=dict(values=dict(tags=[], something_else='loki'), find_me='bingo'),
                        other_key=dict(values=dict(tags=[], something_else='thor'), find_me='b i n g o'))
        search_key = 'tags'
        return_key = 'find_me'
        expected = ['bingo', 'b i n g o']

        self.assertEqual(sorted(find_root_by_key(haystack, search_key, return_key)), sorted(expected))

    def test_are_networks_same_success(self):
        network_a = '192.168.0.0/24'
        networks = ['192.168.0.0/24']
        self.assertTrue(are_networks_same(network_a, networks))

    def test_are_networks_same_false(self):
        network_a = '192.168.0.0/24'
        networks = ['192.168.0.0/23']
        self.assertFalse(are_networks_same(network_a, networks))

    def test_convert_resource_type_success(self):
        self.assertEqual(convert_resource_type('AWS Security Group'), 'aws_security_group')

    def test_convert_resource_type_failure(self):
        self.assertEqual(convert_resource_type('test_resource'), 'test_resource')

    def test_seek_regex_in_dict_value_nested_dict(self):
        haystack = dict(search_key=dict(something='value'))
        key_name = 'something'
        needle = 'val.*'
        expected = ['value']

        self.assertEqual(seek_regex_key_in_dict_values(haystack, key_name, needle), expected)

    def test_jsonify(self):
        self.assertEqual(jsonify({}), {})
        self.assertEqual(jsonify([]), [])
        self.assertEqual(jsonify(12), 12)
        self.assertEqual(jsonify('something'), 'something')
        self.assertEqual(jsonify('{"test": "test_value"}'), {'test': 'test_value'})

    def test_remove_mounted_resources(self, *args):
        resource_list = {
            'address': 'aws_subnet.test',
            'type': 'aws_subnet',
            'name': 'test',
            'values': {
                'tags': None,
                'aws_vpc': [
                    {
                        'tags': {
                            'Environment': 'Sandbox',
                            'Name': 'mcrilly-sandbox'
                        },
                        'aws_subnet': [
                            {
                                'tags': None,
                                'terraform-compliance.mounted': True
                            }
                        ], 'terraform-compliance.mounted': True
                    }
                ]
            },
            'terraform-compliance.mounted_resources': {'key': 'aws_vpc'}
        }
        output = remove_mounted_resources([resource_list])
        self.assertEqual({'tags': None}, output[0]['values'])

    def test_get_resource_name_from_stash_address_exists(self):
        stash = {}
        self.assertEqual({'address': 'test'}, get_resource_name_from_stash(stash=stash, address='test'))

    def test_get_resource_name_from_stash(self):
        stash = [
            {
                'address': 'test'
            }
        ]
        self.assertEqual({'address': 'test'}, get_resource_name_from_stash(stash=stash))

    def test_dict_merge_no_change(self):
        source = {
                'a': [{'a': False}],
                'b': True
        }
        target ={
                'a': [{'a': True}]
        }
        self.assertEqual(dict_merge(source, target), source)

    def test_dict_merge_no_change_1(self):
        source = {
            'a': [{'a': False}],
            'b': True
        }
        target = []
        self.assertEqual(dict_merge(source, target), source)

    def test_dict_merge_no_change_2(self):
        self.assertEqual(dict_merge([], []), [])

    def test_dict_merge_success(self):
        source = {
            'a': [{'a': False}],
            'b': True
        }
        target ={
            'a': [{'b': 0}]
        }
        self.assertEqual(dict_merge(source, target), {'a': [{'a': False}, {'b': 0}], 'b': True})

    def test_is_list_of_dict_failures(self):
        self.assertFalse(is_list_of_dict(['a', 'b', 'c']))
        self.assertFalse(is_list_of_dict([1, 2, 3]))
        self.assertFalse(is_list_of_dict(['a', {'b':'c'}]))
        self.assertFalse(is_list_of_dict([False]))
        self.assertFalse(is_list_of_dict([]))

    def test_is_list_of_dict_successes(self):
        self.assertTrue(is_list_of_dict([{}]))
        self.assertTrue(is_list_of_dict([{'a:': [1]}]))
        self.assertTrue(is_list_of_dict([{'a:': 'b'}]))

    def test_is_key_exist_failures(self):
        self.assertFalse(is_key_exist('key', []))
        self.assertFalse(is_key_exist('key', [{}]))
        self.assertFalse(is_key_exist('key', [{'something': True}]))
        self.assertFalse(is_key_exist('key', [1, 2]))
        self.assertFalse(is_key_exist('key', ['key', 'b']))
        self.assertFalse(is_key_exist('key', [False]))
        self.assertFalse(is_key_exist('key', {'key': []}))

    def test_is_key_exist_successes(self):
        self.assertTrue(is_key_exist('key', [{'key': 'something'}]))
        self.assertTrue(is_key_exist('key', [{'something': True}, 2, 3, 4, {'key': True}]))
        self.assertTrue(is_key_exist('key', [{'something': True}, {'key': True}]))

    def test_transform_asg_style_tags_success(self):
        resource_list = {
            'address': 'aws_subnet.test',
            'type': 'aws_subnet',
            'name': 'test',
            'values': {
                'tags': None,
                'tag': [
                    {
                        'key': 'some_key',
                        'value': 'some_value'
                    }
                ],
            }
        }
        output = transform_asg_style_tags([resource_list])
        self.assertEqual(output[0]['values']['tags']['some_key'], 'some_value')

    def test_transform_asg_style_tags_with_null_values_success(self):
        resource_list = {
            'address': 'aws_subnet.test',
            'type': 'aws_subnet',
            'name': 'test',
            'values': None
        }
        output = transform_asg_style_tags([resource_list])
        self.assertEqual(output[0]['values']['tags'], {})

    def test_transform_asg_style_tags_return_directly(self):
        resource_list = 'abc'
        output = transform_asg_style_tags(resource_list)
        self.assertEqual(output, 'abc')

    def test_python_version_check_success(self):
        self.assertEqual(python_version_check(), 1)

    @data(['module.test["a"].aws_s3_bucket.test["a"]', 'module.test.aws_s3_bucket.test'],
          ['module.test[1].aws_s3_bucket.test["a"]','module.test.aws_s3_bucket.test'],
          ['module.test["a"].aws_s3_bucket.test[2]', 'module.test.aws_s3_bucket.test'],
          ['module.test["a-b"].aws_s3_bucket.test[2]', 'module.test.aws_s3_bucket.test'],
          ['module.test["a_B"].aws_s3_bucket.test[2]', 'module.test.aws_s3_bucket.test'],
          ['module.test["a_53917_xafad"].aws_s3_bucket.test[2]', 'module.test.aws_s3_bucket.test'],
          ['module.test["a_53917/xafad"].aws_s3_bucket.test[2]', 'module.test.aws_s3_bucket.test'],
          ['module.test["a_5391\\7/xafad"].aws_s3_bucket.test[2]', 'module.test.aws_s3_bucket.test'],
          ['module.test.aws_s3_bucket.test', 'module.test.aws_s3_bucket.test'])
    @unpack
    def test_strip_iterations_from_address(self, given, expected):
        self.assertEqual(strip_iterations(given), expected)

    @data(['module.a.module.b.module.c', 'c'],
          ['module.a', 'a'],
          ['module.a.module.b.aws_some_resource.c', 'b'],
          ['aws_some_resource.c', 'aws_some_resource.c'])
    @unpack
    def test_get_most_child_module(self, given, expected):
        self.assertEqual(get_most_child_module(given), expected)



import diskcache
from terraform_compliance.common.defaults import Defaults


class Cache(object):
    def __init__(self):
        self.cache = diskcache.Cache(Defaults.cache_dir)

    def set(self, key, value):
        return self.cache.add(bytes(key, encoding='utf8'), value)

    def get(self, key):
        return self.cache.get(bytes(key, encoding='utf8'))

    def delete(self, key):
        del self.cache[bytes(key, encoding='utf8')]

    def close(self):
        return self.cache.close()



# -*- coding: utf-8 -*-

from radish import world
from terraform_compliance.common.bdd_tags import look_for_bdd_tags
from terraform_compliance.common.helper import (
    get_resource_address_list_from_stash,
    find_root_by_key,
    remove_mounted_resources,
    transform_asg_style_tags,
    convert_resource_type,
)
from terraform_compliance.extensions.ext_radish_bdd import skip_step
import re


def i_have_name_section_configured(_step_obj, name, type_name='resource', _terraform_config=world):
    '''
    Finds given resource or variable by name and returns it. Skips the step (and further steps) if it is not found.

    :param _step_obj: Internal, step object for radish.
    :param name: String of the name of the resource_type or variable.
    :param type_name: String of the type, either resource(s) or variable(s)
    :param _terraform_config: Internal, terraform configuration.
    :return:
    '''
    assert (type_name in ['resource', 'resources',
                          'variable', 'variables',
                          'output', 'outputs',
                          'provider', 'providers',
                          'data', 'datas']), \
        '{} configuration type does not exist or not implemented yet. ' \
        'Use resource(s), provider(s), variable(s), output(s) or data(s) instead.'.format(type_name)

    if type_name.endswith('s'):
        type_name = type_name[:-1]

    # Process the tags
    _step_obj = look_for_bdd_tags(_step_obj)
    match = _step_obj.context.match

    if not hasattr(_step_obj.context, 'cumulative_stash'):
        _step_obj.context.cumulative_stash = []
    if not hasattr(_step_obj.context, 'stash'):
        _step_obj.context.stash = []

    if name in ('a resource', 'any resource', 'resources'):
        _step_obj.context.type = type_name
        _step_obj.context.name = name
        _step_obj.context.stash = [obj for key, obj in _terraform_config.config.terraform.resources_raw.items()]
        _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
        _step_obj.context.addresses = get_resource_address_list_from_stash(_step_obj.context.stash)
        _step_obj.context.property_name = type_name
        return True

    elif name in ('an output', 'any output', 'outputs'):
        _step_obj.context.type = 'output'
        _step_obj.context.name = name
        _step_obj.context.stash = [obj for key, obj in _terraform_config.config.terraform.configuration['outputs'].items()]
        _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
        _step_obj.context.addresses = get_resource_address_list_from_stash(_terraform_config.config.terraform.configuration['outputs'])
        _step_obj.context.property_name = 'output'
        return True

    elif name in ('a variable', 'any variable', 'variables'):
        _step_obj.context.type = 'variable'
        _step_obj.context.name = name
        _step_obj.context.stash = [obj for key, obj in _terraform_config.config.terraform.configuration['variables'].items()]
        _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
        _step_obj.context.addresses = 'variable'
        _step_obj.context.property_name = 'variable'
        return True

    elif name.startswith('resource that supports'):
        filter_property = re.match(r'resource that supports (.*)', name).group(1)

        resource_types_supports_tags = set(find_root_by_key(_terraform_config.config.terraform.resources_raw,
                                                        filter_property,
                                                        return_key='type'))

        # look for after_unknown values in case they were omitted
        for resource_type, after_unknown_properties in _terraform_config.config.terraform.type_to_after_unknown_properties.items():
            if filter_property in after_unknown_properties:
                resource_types_supports_tags.add(resource_type)

        resource_list = []
        for resource_type in resource_types_supports_tags:
            # Issue-168: Mounted resources causes problem on recursive searching for resources that supports tags
            #            We are removing all mounted resources here for future steps, since we don't need them for
            #            tags checking.
            found_resources = remove_mounted_resources(_terraform_config.config.terraform.find_resources_by_type(resource_type, match))
            found_resources = transform_asg_style_tags(found_resources)
            resource_list.extend(found_resources)

        if resource_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = resource_list
            _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
            _step_obj.context.addresses = get_resource_address_list_from_stash(resource_list)
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'resource':
        name = convert_resource_type(name)
        resource_list = _terraform_config.config.terraform.find_resources_by_type(name, match)

        if resource_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = resource_list
            _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
            _step_obj.context.addresses = get_resource_address_list_from_stash(resource_list)
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'variable':
        found_variable = match.get(_terraform_config.config.terraform.variables, name, None)

        if found_variable:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = [found_variable]
            _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
            _step_obj.context.addresses = name
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'output':
        found_output = match.get(_terraform_config.config.terraform.outputs, name, None)

        if found_output:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = found_output
            _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
            _step_obj.context.addresses = name
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'provider':
        found_provider = _terraform_config.config.terraform.get_providers_from_configuration(name, match)

        if found_provider:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = found_provider
            _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
            _step_obj.context.addresses = name
            _step_obj.context.address = name
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'data':
        name = convert_resource_type(name)
        data_list = _terraform_config.config.terraform.find_data_by_type(name, match)

        if data_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = data_list
            _step_obj.context.cumulative_stash.extend(_step_obj.context.stash)
            _step_obj.context.addresses = name
            _step_obj.context.address = name
            _step_obj.context.property_name = type_name
            return True

    skip_step(_step_obj, name)


# -*- coding: utf-8 -*-

from terraform_compliance.common.helper import (
    Null
)
from terraform_compliance.extensions.ext_radish_bdd import skip_step
from terraform_compliance.common.error_handling import Error
from terraform_compliance.main import Step
from radish.utils import console_write
from terraform_compliance.common.defaults import Defaults


def it_contains_something_old(_step_obj, something, inherited_values=Null):
    console_write("\t{} {}: {}".format(Defaults().warning_icon,
                                       Defaults().warning_colour('WARNING'),
                                       Defaults().info_colour('"When it contains {}" step functionality will be changed'
                                                              ' on future versions and the functionality will be same '
                                                              'as "When it has {}" step. Please use the '
                                                              'latter.'.format(something, something))))
    match = _step_obj.context.match
    seek_key_in_dict = match.seek_key_in_dict
    seek_regex_key_in_dict_values = match.seek_regex_key_in_dict_values

    prop_list = []

    _step_obj.context.stash = inherited_values if inherited_values is not Null else _step_obj.context.stash

    if _step_obj.context.type in ('resource', 'data'):
        for resource in _step_obj.context.stash:
            if not isinstance(resource, dict) \
                    or 'values' not in resource \
                    or 'address' not in resource \
                    or 'type' not in resource:
                resource = {'values': resource,
                            'address': resource,
                            'type': _step_obj.context.name}

            values = resource.get('values', resource.get('expressions', {}))
            if not values:
                values = seek_key_in_dict(resource, something)

            found_value = Null
            found_key = Null
            if isinstance(values, dict):
                found_key = match.get(values, something, seek_key_in_dict(values, something))
                if not isinstance(found_key, list):
                    found_key = [{something: found_key}]

                if len(found_key):
                    found_key = found_key[0] if len(found_key) == 1 and match.contains(found_key[0], something) else found_key

                    if isinstance(found_key, dict):
                        found_value = match.get(found_key, something, found_key)
                    else:
                        found_value = found_key
            elif isinstance(values, list):
                found_value = []

                for value in values:

                    if isinstance(value, dict):
                        # First search in the keys
                        found_key = seek_key_in_dict(value, something)

                        # Then search in the values with 'key'
                        if not found_key:
                            found_key = seek_regex_key_in_dict_values(value, 'key', something)

                            if found_key:
                                found_key = found_key[0]
                                found_value = value.get('value')
                                break
                    elif isinstance(value, list):
                        found_key, found_value = it_contains_something_old(_step_obj, something, value)

                    if found_key is not Null and len(found_key):
                        found_key = found_key[0] if len(found_key) == 1 else found_key

                        if isinstance(found_key, dict):
                            found_value.append(match.get(found_key, something, found_key))

            if isinstance(found_value, dict) and 'constant_value' in found_value:
                found_value = found_value['constant_value']

            if found_value is not Null and found_value != [] and found_value != '' and found_value != {}:
                prop_list.append({'address': resource['address'],
                                  'values': found_value,
                                  'type': _step_obj.context.name})

        if prop_list:
            _step_obj.context.stash = prop_list
            _step_obj.context.property_name = something

            return something, prop_list

        if _step_obj.state != Step.State.FAILED:
            skip_step(_step_obj,
                      resource=_step_obj.context.name,
                      message='Can not find any {} property for {} resource in '
                              'terraform plan.'.format(something, _step_obj.context.name))

    elif _step_obj.context.type == 'provider':
        for provider_data in _step_obj.context.stash:
            values = seek_key_in_dict(provider_data, something)

            if values:
                _step_obj.context.stash = values
                _step_obj.context.property_name = something
                _step_obj.context.address = '{}.{}'.format(provider_data.get('name', _step_obj.context.addresses),
                                                           provider_data.get('alias', "\b"))
                return True

    if _step_obj.state != Step.State.FAILED:
        skip_step(_step_obj,
                  resource=_step_obj.context.name,
                  message='Skipping the step since {} type does not have {} property.'.format(_step_obj.context.type,
                                                                                              something))


def it_has_something(_step_obj, something, inherited_values=Null):
    match = _step_obj.context.match
    seek_key_in_dict = match.seek_key_in_dict
    seek_regex_key_in_dict_values = match.seek_regex_key_in_dict_values

    prop_list = []

    _step_obj.context.stash = inherited_values if inherited_values is not Null else _step_obj.context.stash

    if _step_obj.context.type in ('resource', 'data'):
        for resource in _step_obj.context.stash:
            if not isinstance(resource, dict) \
                    or 'values' not in resource \
                    or 'address' not in resource \
                    or 'type' not in resource:
                resource = {'values': resource,
                            'address': resource,
                            'type': _step_obj.context.name}

            values = resource.get('values', resource.get('expressions', {}))
            if not values:
                values = seek_key_in_dict(resource, something)

            found_value = Null
            found_key = Null
            if isinstance(values, dict):
                found_key = match.get(values, something, Null)
                if found_key is not Null:
                    found_key = [{something: found_key}]
                else:
                    found_key = seek_key_in_dict(values, something)

                if len(found_key):
                    found_key = found_key[0] if len(found_key) == 1 and match.contains(found_key[0], something) else found_key

                    if isinstance(found_key, dict):
                        found_value = match.get(found_key, something, found_key)
                    else:
                        found_value = found_key
            elif isinstance(values, list):
                found_value = []

                for value in values:

                    if isinstance(value, dict):
                        # First search in the keys
                        found_key = seek_key_in_dict(value, something)

                        # Then search in the values with 'key'
                        if not found_key:
                            found_key = seek_regex_key_in_dict_values(value, 'key', something)

                            if found_key:
                                found_key = found_key[0]
                                found_value = value.get('value')
                                break
                    elif isinstance(value, list):
                        found_key, found_value = it_has_something(_step_obj, something, value)
                    elif isinstance(value, (str, bool, int, float)) and match.equals(value, something):
                        found_key = value
                        found_value.append(found_key)

                    if found_key is not Null and len(found_key):
                        found_key = found_key[0] if len(found_key) == 1 else found_key

                        if isinstance(found_key, dict):
                            found_value.append(match.get(found_key, something, found_key))

            if isinstance(found_value, dict) and 'constant_value' in found_value:
                found_value = found_value['constant_value']

            if found_value not in (Null, [], '', {}):
                prop_list.append(resource)

        if prop_list:
            _step_obj.context.stash = prop_list
            _step_obj.context.property_name = something

            return something, prop_list

        if _step_obj.state != Step.State.FAILED:
            skip_step(_step_obj,
                      resource=_step_obj.context.name,
                      message='Can not find any {} property for {} resource in '
                              'terraform plan.'.format(something, _step_obj.context.name))

    elif _step_obj.context.type == 'provider':
        for provider_data in _step_obj.context.stash:
            values = seek_key_in_dict(provider_data, something)

            if values:
                prop_list.extend(values)
                _step_obj.context.property_name = something
                _step_obj.context.address = '{}.{}'.format(provider_data.get('name', _step_obj.context.addresses),
                                                           provider_data.get('alias', "\b"))

        if prop_list:
            _step_obj.context.stash = prop_list
            return True

    if _step_obj.state != Step.State.FAILED:
        skip_step(_step_obj,
                  resource=_step_obj.context.name,
                  message='Skipping the step since {} type does not have {} property.'.format(_step_obj.context.type,
                                                                                              something))


def it_does_not_have_something(_step_obj, something, inherited_values=Null):
    match = _step_obj.context.match
    seek_key_in_dict = match.seek_key_in_dict
    seek_regex_key_in_dict_values = match.seek_regex_key_in_dict_values

    prop_list = []

    _step_obj.context.stash = inherited_values if inherited_values is not Null else _step_obj.context.stash

    if _step_obj.context.type in ('resource', 'data'):
        for resource in _step_obj.context.stash:
            if not isinstance(resource, dict) \
                    or 'values' not in resource \
                    or 'address' not in resource \
                    or 'type' not in resource:
                resource = {'values': resource,
                            'address': resource,
                            'type': _step_obj.context.name}

            values = resource.get('values', resource.get('expressions', {}))
            if not values:
                values = seek_key_in_dict(resource, something)

            found_value = Null
            found_key = Null
            if isinstance(values, dict):
                found_key = match.get(values, something, Null)
                if found_key is not Null:
                    found_key = [{something: found_key}]
                else:
                    found_key = seek_key_in_dict(values, something)

                if len(found_key):
                    found_key = found_key[0] if len(found_key) == 1 and match.contains(found_key[0], something) else found_key

                    if isinstance(found_key, dict):
                        found_value = match.get(found_key, something, found_key)
                    else:
                        found_value = found_key
            elif isinstance(values, list):
                found_value = []

                for value in values:

                    if isinstance(value, dict):
                        # First search in the keys
                        found_key = seek_key_in_dict(value, something)

                        # Then search in the values with 'key'
                        if not found_key:
                            found_key = seek_regex_key_in_dict_values(value, 'key', something)

                            if found_key:
                                found_key = found_key[0]
                                found_value = value.get('value')
                                break
                    elif isinstance(value, list):
                        found_key, found_value = it_has_something(_step_obj, something, value)

                    if found_key is not Null and len(found_key):
                        found_key = found_key[0] if len(found_key) == 1 else found_key

                        if isinstance(found_key, dict):
                            found_value.append(match.get(found_key, something, found_key))

            if isinstance(found_value, dict) and 'constant_value' in found_value:
                found_value = found_value['constant_value']

            if found_value is not Null and found_value != [] and found_value != '' and found_value != {}:
                prop_list.append(resource['address'])


        prop_list = [resource for resource in _step_obj.context.stash if resource['address'] not in prop_list]
        _step_obj.context.property_name = something

        if prop_list:
            _step_obj.context.stash = prop_list
            return something, prop_list

        if _step_obj.state != Step.State.FAILED:
            skip_step(_step_obj,
                      resource=_step_obj.context.name,
                      message='All objects ({}) coming from previous step has {} '
                              'property.'.format(_step_obj.context.name, something))

    elif _step_obj.context.type == 'provider':
        stash = []
        for provider_data in _step_obj.context.stash:
            values = seek_key_in_dict(provider_data, something)

            if values:
                return False
            else:
                stash.append(provider_data)

        if stash:
            _step_obj.context.stash = stash
            _step_obj.context.property_name = something
            return True


    return True



from terraform_compliance.common.exceptions import Failure
from radish import world
from radish.utils import console_write
from terraform_compliance.main import Step
import colorful
from ast import literal_eval
from mock import MagicMock
import os


class WrapperError(Exception):
    def __init__(self, exception):
        self.exception = exception
        self.reason = str(exception)
        self.traceback = "No Traceback"
        self.name = exception.__class__.__name__
        self.filename = None
        self.line = 0


# This class only works with steps
class Error(Exception):
    def __init__(self, step_obj, message, exception=Failure):
        self.message = message.split("\n")
        if isinstance(world.config.user_data['exit_on_failure'], bool):
            self.exit_on_failure = world.config.user_data['exit_on_failure']
            self.no_failure = world.config.user_data['no_failure']
        elif isinstance(world.config.user_data, MagicMock):
            self.exit_on_failure = True
            self.no_failure = False
        else:
            self.exit_on_failure = literal_eval(world.config.user_data['exit_on_failure'])
            self.no_failure = literal_eval(world.config.user_data['no_failure'])
        _TFC_ERROR = os.environ.get('TFC_ERROR')

        if step_obj.context.no_failure:
            self.exception = type(step_obj.context.failure_class, (Exception, ), {})
            self.exception_name = step_obj.context.failure_class
            self.no_failure = True
        elif _TFC_ERROR is None:
            self.exception = exception
            self.exception_name = exception.__name__
        else:
            self.exception = type(_TFC_ERROR, (Exception, ), {})
            self.exception_name = _TFC_ERROR

        self.step_obj = step_obj
        self._process()

    def _process(self):
        # Prepare message
        msg = []
        for msg_index in range(0,len(self.message)):
            if self.exit_on_failure is False or self.no_failure is True:
                msg_header = '{}{}'.format(self.exception_name,
                                           colorful.bold_white(':')) if msg_index == 0 else ' '*(len(self.exception_name)+1)
                if str(world.config.formatter) in ('gherkin'):
                    # this line could be improved by letting radish handle the printing
                    msg.append('\t\t{} {}'.format(colorful.bold_red(msg_header), colorful.red(self.message[msg_index])))
                elif str(world.config.formatter) in ('silent_formatter'):
                    msg.append('{} '.format(colorful.bold_red(msg_header)))
                    msg.append('{}'.format(colorful.red(self.message[msg_index])))

            else:
                msg.append(self.message[msg_index] if msg_index == 0
                                                   else '{}{} {} {}'.format("\t"*2,
                                                                            ' '*(len(self.exception_name)+1),
                                                                            colorful.bold_white(':'),
                                                                            self.message[msg_index]))

        if self.exit_on_failure is False or (self.no_failure is True and msg):

            if str(world.config.formatter) in ('gherkin'):
                for message in msg:
                    console_write(message)
            elif str(world.config.formatter) in ('silent_formatter'):
                if not hasattr(self.step_obj.context, 'failure_msg'): # where to put this
                    self.step_obj.context.failure_msg = []
                self.step_obj.context.failure_msg.extend(msg)


            if self.no_failure is False:
                self._fail_step(self.step_obj.id)
            else:
                self.step_obj.state = Step.State.SKIPPED
                for step in self.step_obj.parent.all_steps:
                    step.runable = False
            return

        if self.no_failure is False:
            raise self.exception('\n'.join(msg))

    def _fail_step(self, step_id):
        for step in self.step_obj.parent.all_steps:
            if step.id == step_id:
                step.state = Step.State.FAILED
                step.failure = WrapperError(self.exception('\r{}'.format(' '*len(self.exception_name))))

        if self.message:
            self.step_obj.failure.traceback = '{}: {}'.format(self.exception_name, '\n'.join(self.message)).replace('\x00', '\\x00').replace('\x08', '\\x08')



import sys
import os
from argparse import Action
import filetype
from terraform_compliance.common.terraform_files import convert_terraform_plan_to_json
from terraform_compliance.common.exceptions import TerraformComplianceInternalFailure

try:
    import orjson as json
except ImportError:
    import json


class ReadablePlan(Action):
    def __init__(self, option_strings, dest, **kwargs):
        super(ReadablePlan, self).__init__(option_strings, dest, **kwargs)
        self.option_strings = option_strings
        self.dest = dest
        self.nargs = kwargs.get('nargs', None)
        self.const = kwargs.get('const', None)
        self.default = kwargs.get('default', None)
        self.type = kwargs.get('type', None)
        self.choices = kwargs.get('choices', None)
        self.required = kwargs.get('required', None)
        self.help = kwargs.get('help', None)
        self.metavar = kwargs.get('metavar', None)
        self.parser = kwargs.get('parser', None)

    def __call__(self, parser, namespace, values, option_string=None):
        # Check if the given path is a file
        if not os.path.isfile(values):
            print('ERROR: {} is not a file.'.format(values))
            sys.exit(1)

        # Check if the given file is a native terraform plan file
        given_file = filetype.guess(values)

        if given_file is not None:
            terraform_executable = getattr(namespace, 'terraform_file', None)
            values = convert_terraform_plan_to_json(os.path.abspath(values), terraform_executable)

        # Check if the given file is a json file.
        try:
            plan_lines = [line for line in open(values, 'r', encoding='utf-8')]

            # Some Github Actions (hashicorp/setup-terraform) has internal wrappers which is
            # breaking the json file that is read by the terraform-compliance
            file_change_required = False
            if len(plan_lines) > 1:
                plan_lines = plan_lines[1]
                file_change_required = True
            else:
                plan_lines = plan_lines[0]

            data = json.loads(plan_lines)

            # Write the changed plan file to the same file, since it is used in other places.
            if file_change_required:
                with open(values, 'w', encoding='utf-8') as plan_file:
                    plan_file.write(plan_lines)

        except json.JSONDecodeError:
            print('ERROR: {} is not a valid JSON file'.format(values))
            sys.exit(1)
        except UnicodeDecodeError:
            print('ERROR: {} is not a valid JSON file.'.format(values))
            print('       Did you try to convert the binary plan file to json with '
                  '"terraform show -json {} > {}.json" ?'.format(values, values))
            sys.exit(1)
        except:
            raise TerraformComplianceInternalFailure('Invalid file type.')

        # Check if this is a correct terraform plan file
        try:
            assert data['format_version']
            assert data['terraform_version']

            # Check if this is a state file
            if 'values' in data:
                assert data['values']['root_module'].get('resources', data['values']['root_module'].get('child_modules'))

            # Then it must be a terraform plan file
            else:
                assert data['planned_values']
                assert data['configuration']

        except KeyError:
            print('ERROR: {} is not a valid terraform plan json output.'.format(values))
            sys.exit(1)

        setattr(namespace, self.dest, os.path.abspath(values))
        return True



from unittest import TestCase
from terraform_compliance.extensions.ext_radish_bdd import custom_type_any, custom_type_section


class Test_Steps_Custom_Type(TestCase):

    def test_custom_type_any(self):
        self.assertEqual(custom_type_any('AnyType'), 'AnyType')
        self.assertEqual(custom_type_any('13/135lx.13-_f19  39'), '13/135lx.13-_f19  39')

    def test_custom_type_section(self):
        for section in ['resource', 'provider', 'data', 'variable']:
            self.assertEqual(custom_type_section(section), section)

        self.assertIsNone(custom_type_section('this is something else'))


from terraform_compliance.common.error_handling import Error
from terraform_compliance.common.helper import Null
import re


def its_value_condition_be_null(_step_obj, condition):
    if condition not in ['must', 'must not']:
        Error(_step_obj, 'Condition must be one of "must" or "must not"')
        return

    condition_must = condition == 'must'

    match = _step_obj.context.match
    regex = r'{}'.format(u'(\x00|^$)$')
    is_null = lambda x: x in ([], {}, None, Null) or (isinstance(x, str) and match.regex_match(regex, x))

    for resource in _step_obj.context.stash:
        values = resource.get('values', {})

        values_is_null = is_null(values)
        if (values_is_null and not condition_must) or (not values_is_null and condition_must):
            Error(_step_obj, '{} property in {} is considered to be {}. It is set to "{}".'.format(_step_obj.context.property_name,
                                                                                                   resource.get('address', _step_obj.context.name),
                                                                                                   'Null' if values_is_null else 'not Null',
                                                                                                   values))


        elif isinstance(values, (dict, list)):

            list_of_inner_values = values.values() if isinstance(values, dict) else values
            # all values in values are considered null
            all_null = all(map(is_null, list_of_inner_values))

            if (all_null and not condition_must) or (not all_null and condition_must):
                Error(_step_obj, 'Multivalued {} property in {} is considered to be {}. It is set to "{}".'.format(_step_obj.context.property_name,
                                                                                                                   resource.get('address', _step_obj.context.name),
                                                                                                                   'Null' if all_null else 'not Null',
                                                                                                                   values))




# -*- coding: utf-8 -*-

from radish import (
    world,
    given,
    when,
    then
)
from terraform_compliance.common.helper import (
    Null,
    EmptyStash
)
import re
from terraform_compliance.common.error_handling import Error

# GIVEN
from terraform_compliance.steps.given.i_have_name_section_configured import i_have_name_section_configured

# WHEN
from terraform_compliance.steps.when.its_key_metadata_has_something import its_key_metadata_has_something
from terraform_compliance.steps.when.its_key_is_value import (
    its_key_is_value,
    its_key_is_not_value
)
from terraform_compliance.steps.when.maths import (
    i_action_them,
    i_expect_the_result_is_operator_than_number
)
from terraform_compliance.steps.when.it_contains_something import (
    it_contains_something_old,
    it_has_something,
    it_does_not_have_something
)

# THEN
from terraform_compliance.steps.then.it_must_contain_something import (
    it_must_contain_something,
    it_must_not_contain_something,
)
from terraform_compliance.steps.then.property_is_enabled import property_is_enabled
from terraform_compliance.steps.then.security_group_related import it_condition_have_proto_protocol_and_port_port_for_cidr
from terraform_compliance.steps.then.its_value_condition_match_the_search_regex import its_value_condition_match_the_search_regex_regex
from terraform_compliance.steps.then.its_value_condition_match_the_search_regex import its_singular_value_condition_match_the_search_regex_regex
from terraform_compliance.steps.then.its_value_condition_match_the_search_regex import any_of_its_values_condition_match_the_search_regex_regex
from terraform_compliance.steps.then.its_value_condition_contain import its_value_condition_contain
from terraform_compliance.steps.then.it_must_have_reference_address_referenced import it_must_have_reference_address_referenced
from terraform_compliance.steps.then.its_key_condition_be_value import its_key_condition_be_value
from terraform_compliance.steps.then.interpolations import i_flatten_everything_found
from terraform_compliance.steps.then.it_must_be_in import it_must_be_in
from terraform_compliance.steps.then.it_must_not_be_in import it_must_not_be_in
from terraform_compliance.steps.then.its_value_condition_be_null import its_value_condition_be_null

# {name} is checked for startswith("resource that supports "). The @given decorator for that documented case is not needed
@given(u'I have {name:ANY} defined')
@given(u'I have {name:ANY} {type_name:SECTION} configured')
def wrapper(_step_obj, name, type_name='resource', _terraform_config=world):
    return i_have_name_section_configured(_step_obj, name, type_name, _terraform_config)


@when(u'its {key:PROPERTY} metadata has {value:PROPERTY}')
@when(u'its {key:PROPERTY} metadata is {value:PROPERTY}')
@when(u'its {key:PROPERTY} metadata contains {value:PROPERTY}')
@when(u'its {key:PROPERTY} metadata includes {value:PROPERTY}')
def wrapper(_step_obj, key, value):
    return its_key_metadata_has_something(_step_obj, key, value, regex_match=False)

@when(u'its {key:PROPERTY} metadata has "{search_regex}" regex')
@when(u'its {key:PROPERTY} metadata is "{search_regex}" regex')
@when(u'its {key:PROPERTY} metadata contains "{search_regex}" regex')
@when(u'its {key:PROPERTY} metadata includes "{search_regex}" regex')
def wrapper(_step_obj, key, search_regex):
    return its_key_metadata_has_something(_step_obj, key, search_regex, regex_match=True)

@when(u'its {key:PROPERTY} metadata has not {value:PROPERTY}')
@when(u'its {key:PROPERTY} metadata is not {value:PROPERTY}')
@when(u'its {key:PROPERTY} metadata does not have {value:PROPERTY}')
@when(u'its {key:PROPERTY} metadata does not contain {value:PROPERTY}')
@when(u'its {key:PROPERTY} metadata does not include {value:PROPERTY}')
def wrapper(_step_obj, key, value):
    return its_key_metadata_has_something(_step_obj, key, value, has_step=False)

@when(u'its {key:PROPERTY} metadata has not "{search_regex}" regex')
@when(u'its {key:PROPERTY} metadata is not "{search_regex}" regex')
@when(u'its {key:PROPERTY} metadata does not have "{search_regex}" regex')
@when(u'its {key:PROPERTY} metadata does not contain "{search_regex}" regex')
@when(u'its {key:PROPERTY} metadata does not include "{search_regex}" regex')
def wrapper(_step_obj, key, search_regex):
    return its_key_metadata_has_something(_step_obj, key, search_regex, has_step=False, regex_match=True)

@when(u'its {key:PROPERTY} is {value:PROPERTY}')
@when(u'its {key:PROPERTY} has {value:PROPERTY}')
@when(u'its {key:PROPERTY} includes {value:PROPERTY}')
@when(u'its {key:PROPERTY} contains {value:PROPERTY}')
@when(u'its {address:PROPERTY} {key:PROPERTY} is {value:PROPERTY}')
@when(u'its {address:PROPERTY} {key:PROPERTY} has {value:PROPERTY}')
@when(u'its {address:PROPERTY} {key:PROPERTY} includes {value:PROPERTY}')
@when(u'its {address:PROPERTY} {key:PROPERTY} contains {value:PROPERTY}')
@when(u'its {key:PROPERTY} includes an entry where {value:PROPERTY} is {dict_value:PROPERTY}')
def wrapper(_step_obj, key, value, dict_value=None, address=Null):
    return its_key_is_value(_step_obj, key, value, dict_value, address, regex_match=False)

@when(u'its {key:PROPERTY} is "{search_regex}" regex')
@when(u'its {key:PROPERTY} has "{search_regex}" regex')
@when(u'its {key:PROPERTY} includes "{search_regex}" regex')
@when(u'its {key:PROPERTY} contains "{search_regex}" regex')
@when(u'its {address:PROPERTY} {key:PROPERTY} is "{search_regex}" regex')
@when(u'its {address:PROPERTY} {key:PROPERTY} has "{search_regex}" regex')
@when(u'its {address:PROPERTY} {key:PROPERTY} includes "{search_regex}" regex')
@when(u'its {address:PROPERTY} {key:PROPERTY} contains "{search_regex}" regex')
@when(u'its {key:PROPERTY} includes an entry where {value:PROPERTY} is "{dict_value}" regex')
def wrapper(_step_obj, key, search_regex, dict_value=None, address=Null):
    return its_key_is_value(_step_obj, key, search_regex, dict_value, address, regex_match=True)


@when(u'its {key:PROPERTY} is not {value:PROPERTY}')
@when(u'its {key:PROPERTY} has not {value:PROPERTY}')
@when(u'its {key:PROPERTY} does not include {value:PROPERTY}')
@when(u'its {key:PROPERTY} does not contain {value:PROPERTY}')
@when(u'its {address:PROPERTY} {key:PROPERTY} is not {value:PROPERTY}')
@when(u'its {address:PROPERTY} {key:PROPERTY} has not {value:PROPERTY}')
@when(u'its {address:PROPERTY} {key:PROPERTY} does not include {value:PROPERTY}')
@when(u'its {address:PROPERTY} {key:PROPERTY} does not contain {value:PROPERTY}')
@when(u'its {key:PROPERTY} does not include an entry where {value:PROPERTY} is {dict_value:PROPERTY}')
def wrapper(_step_obj, key, value, dict_value=None, address=Null):
    return its_key_is_not_value(_step_obj, key, value, dict_value, address, regex_match=False)

@when(u'its {key:PROPERTY} is not "{search_regex}" regex')
@when(u'its {key:PROPERTY} has not "{search_regex}" regex')
@when(u'its {key:PROPERTY} does not include "{search_regex}" regex')
@when(u'its {key:PROPERTY} does not contain "{search_regex}" regex')
@when(u'its {address:PROPERTY} {key:PROPERTY} is not "{search_regex}" regex')
@when(u'its {address:PROPERTY} {key:PROPERTY} has not "{search_regex}" regex')
@when(u'its {address:PROPERTY} {key:PROPERTY} does not include "{search_regex}" regex')
@when(u'its {address:PROPERTY} {key:PROPERTY} does not contain "{search_regex}" regex')
@when(u'its {key:PROPERTY} does not include an entry where {value:PROPERTY} is "{dict_value}" regex')
def wrapper(_step_obj, key, search_regex, dict_value=None, address=Null):
    return its_key_is_not_value(_step_obj, key, search_regex, dict_value, address, regex_match=True)

@when(u'it contain {something:PROPERTY}') # This is just here for not breaking backward compatibility. I know its wrong.
@when(u'it contains {something:PROPERTY}')
def wrapper(_step_obj, something, inherited_values=Null):
    return it_contains_something_old(_step_obj, something, inherited_values=Null)


@when(u'it has {something:PROPERTY}')
@when(u'they have {something:PROPERTY}')
def wrapper(_step_obj, something, inherited_values=Null):
    return it_has_something(_step_obj, something, inherited_values=Null)


@when(u'it does not contain {something:PROPERTY}')
@when(u'it doesnt contain {something:PROPERTY}')
@when(u'they do not have {something:PROPERTY}')
@when(u'they dont have {something:PROPERTY}')
@when(u'it does not have {something:PROPERTY}')
@when(u'it doesnt have {something:PROPERTY}')
def wrapper(_step_obj, something, inherited_values=Null):
    return it_does_not_have_something(_step_obj, something, inherited_values=Null)


@then(u'it must have "{reference_address}" referenced')
def wrapper(_step_obj, reference_address):
    return it_must_have_reference_address_referenced(_step_obj, reference_address)


@then(u'it {condition:ANY} have {proto} protocol and port {port} for {cidr:ANY}')
def wrapper(_step_obj, condition, proto, port, cidr):
    return it_condition_have_proto_protocol_and_port_port_for_cidr(_step_obj, condition, proto, port, cidr)


@then(u'it must contain {something:PROPERTY_COMPAT}')
@then(u'it must have {something:PROPERTY_COMPAT}')
@then(u'they must contain {something:PROPERTY_COMPAT}')
@then(u'they must have {something:PROPERTY_COMPAT}')
def wrapper(_step_obj, something, inherited_values=Null):
    return it_must_contain_something(_step_obj, something, inherited_values=Null)


@then(u'it must not contain {something:PROPERTY_COMPAT}')
@then(u'they must not contain {something:PROPERTY_COMPAT}')
@then(u'it must not have {something:PROPERTY_COMPAT}')
@then(u'they must not have {something:PROPERTY_COMPAT}')
def wrapper(_step_obj, something, inherited_values=Null):
    return it_must_not_contain_something(_step_obj, something, inherited_values=Null)


@then(u'{something:ANY} is enabled')
@then(u'{something:ANY} is be enabled')
@then(u'{something:ANY} must be enabled')
def wrapper(_step_obj, something):
    return property_is_enabled(_step_obj, something)


@when(u'I {action_type:PROPERTY} it')
@when(u'I {action_type:PROPERTY} them')
@when(u'I {action_type:PROPERTY} the value')
def wrapper(_step_obj, action_type):
    return i_action_them(_step_obj, action_type)


@then(u'its value must be {operator:ANY} than {number:d}')
@then(u'I expect the result is {operator:ANY} than {number:d}')
@then(u'its value must be {operator:ANY} to {number:d}')
@then(u'I expect the result is {operator:ANY} to {number:d}')
def wrapper(_step_obj, operator, number, _stash=EmptyStash):
    return i_expect_the_result_is_operator_than_number(_step_obj, operator, number, _stash=EmptyStash)


@then(u'its value {condition:ANY} match the "{search_regex}" regex')
@then(u'all of its values {condition:ANY} match the "{search_regex}" regex')
def wrapper(_step_obj, condition, search_regex, _stash=EmptyStash):
    return its_value_condition_match_the_search_regex_regex(_step_obj,
                                                                condition,
                                                                search_regex,
                                                                _stash=EmptyStash)


@then(u'its singular value {condition:ANY} match the "{search_regex}" regex')
def wrapper(_step_obj, condition, search_regex, _stash=EmptyStash):
    return its_singular_value_condition_match_the_search_regex_regex(_step_obj,
                                                                condition,
                                                                search_regex,
                                                                _stash=EmptyStash)


@then(u'any of its values {condition:ANY} match the "{search_regex}" regex')
def wrapper(_step_obj, condition, search_regex, _stash=EmptyStash):
    return any_of_its_values_condition_match_the_search_regex_regex(_step_obj,
                                                                condition,
                                                                search_regex,
                                                                _stash=EmptyStash)


@then(u'its value {condition:ANY} be null')
def wrapper(_step_obj, condition):
    return its_value_condition_be_null(_step_obj, condition)


@then(u'its value {condition:ANY} be {match:ANY}')
def its_value_condition_equal(_step_obj, condition, match, _stash=EmptyStash):
    if match not in ('null'):
        return its_value_condition_match_the_search_regex_regex(_step_obj,
                                                                condition,
                                                                "^" + re.escape(match) + "$",
                                                                _stash)


@then(u'its value {condition:ANY} contain {value:ANY}')
def wrapper(_step_obj, condition, value, _stash=EmptyStash):
    return its_value_condition_contain(_step_obj, condition, value, _stash=EmptyStash)


@then(u'the scenario fails')
@then(u'the scenario should fail')
@then(u'the scenario must fail')
@then(u'it fails')
@then(u'it should fail')
@then(u'it must fail')
def it_fails(_step_obj):
    Error(_step_obj, 'Forcefully failing the scenario on {} ({}) {}'.format(_step_obj.context.name,
                                                                            ', '.join(_step_obj.context.addresses),
                                                                            _step_obj.context.type))


@then('its {key:ANY} {condition:ANY} be {value:ANY}')
@then('its {key:ANY} property {condition:ANY} be {value:ANY}')
@then('its {key:ANY} key {condition:ANY} be {value:ANY}')
def wrapper(_step_obj, key, condition, value, stash=Null, depth=0):
    return its_key_condition_be_value(_step_obj, key, condition, value, stash=Null, depth=0)


@then('I flatten all values found')
def wrapper(_step_obj):
    return i_flatten_everything_found(_step_obj)

@then('it must be a subset of {haystack:ANY}')
@then('it must be in {haystack:ANY}')
def wrapper(_step_obj, haystack):
    return it_must_be_in(_step_obj, haystack)

@then('it must be a superset of {haystack:ANY}')
@then('it must cover {haystack:ANY}')
def wrapper(_step_obj, haystack):
    return it_must_be_in(_step_obj, haystack, cover=True)

@then('it must not be a subset of {haystack:ANY}')
@then('it must not be in {haystack:ANY}')
def wrapper(_step_obj, haystack):
    return it_must_not_be_in(_step_obj, haystack)

@then('it must not be a superset of {haystack:ANY}')
@then('it must not cover {haystack:ANY}')
def wrapper(_step_obj, haystack):
    return it_must_not_be_in(_step_obj, haystack, cover=True)


# -*- coding: utf-8 -*-

from terraform_compliance.common.helper import (
    EmptyStash,
    get_resource_name_from_stash
)
from terraform_compliance.common.error_handling import Error
from terraform_compliance.common.exceptions import TerraformComplianceNotImplemented
from terraform_compliance.common.exceptions import TerraformComplianceInternalFailure


def its_value_condition_contain(_step_obj, condition, value, _stash=EmptyStash):
    match = _step_obj.context.match
    seek_value_in_dict = match.seek_value_in_dict

    if condition not in ('must', 'must not'):
        raise TerraformComplianceNotImplemented('Condition should be one of: `must`, `must not`')

    values = _step_obj.context.stash if _stash is EmptyStash else _stash
    # TODO: Update here for checking values in a list or dict.

    if isinstance(values, list):
        for elem in values:
            values = its_value_condition_contain(_step_obj, condition, value, elem)

    if isinstance(values, (int, bool, str, float)):
        values = dict(values=values,
                      address=_step_obj.context.address if hasattr(_step_obj.context, 'address') else _step_obj.context.addresses)

    found_values = seek_value_in_dict(value, values)
    condition = condition == 'must'

    if condition and not found_values:
        if isinstance(values, list):
            objects = []
            for elem in values:
                objects.append(elem.get('address', '???'))
            objects = ', '.join(objects)
        else:
            objects = values.get('address')

        Error(_step_obj, '{} could not found in {}.'.format(value, objects))

    elif not condition and found_values:
        Error(_step_obj, '{} found in {}.'.format(value,
                                                  get_resource_name_from_stash(found_values).get('address')))

    return values


def _its_value_condition_contain(_step_obj, condition, value, values):
    if isinstance(values, list):
        values = [str(v) for v in values]
        _fail_text = 'did not contain' if condition == 'must' else 'contains'
        fail_message = '{} property in {} {} {}, it is set to {}'.format(
            _step_obj.context.property_name,
            _step_obj.context.name,
            _fail_text,
            value,
            values,
        )
        if condition == 'must':
            assert value in values, fail_message
        else:
            assert value not in values, fail_message
    else:
        raise TerraformComplianceInternalFailure('Can only check that if list contains value')



import os
import stat
import sys
import subprocess
import platform
import urllib.request
import tempfile
from shutil import unpack_archive
import re


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

            windows_executable = os.path.join(path, '{}.exe'.format(program))
            if is_exe(windows_executable):
                return windows_executable
    return None


def convert_terraform_plan_to_json(terraform_plan_file, terraform_executable=None):
    print('. Converting terraform plan file.')
    if terraform_executable is None:
        terraform_executable = which('terraform')
    else:
        print('. Using {} as terraform executable.'.format(terraform_executable))

    if terraform_executable is None:
        sys.stderr.write('ERROR: Could not find "terraform" executable in PATH. Please either use "-t" parameter '
                         'or add terraform executable to your PATH.\n')
        sys.exit(1)

    path = os.path.dirname(terraform_plan_file)
    cwd = os.getcwd()
    os.chdir(path)

    try:
        with open('{}.json'.format(terraform_plan_file), 'w', encoding='utf-8') as FP_plan_file:
            terraform = subprocess.run([terraform_executable, 'show', '-json', terraform_plan_file],
                                       universal_newlines=True,
                                       stdout=FP_plan_file,
                                       stderr=subprocess.PIPE)
    except FileNotFoundError as err:
        sys.stderr.write(
            'ERROR: {} does not exist. Please give correct executable for "terraform".\n'.format(terraform_executable))
        sys.stderr.write('       {}\n'.format(str(err)))
        sys.exit(1)
    except PermissionError  as err:
        sys.stderr.write('ERROR: {} is not executable. Please give correct executable for "terraform".\n'.format(
            terraform_executable))
        sys.stderr.write('       {}\n'.format(str(err)))
        sys.exit(1)
    except OSError as err:
        sys.stderr.write(
            'ERROR: {} does not look like terraform. Please give correct executable for "terraform".\n'.format(
                terraform_executable))
        sys.stderr.write('       {}\n'.format(str(err)))
        sys.exit(1)

    os.chdir(cwd)

    if terraform.returncode == 0:
        return '{}.json'.format(terraform_plan_file)

    # Ok here we are going to fail because plan file has been created by a different terraform
    # So, we will attempt a detection of terraform version here. Hoping that the terraform stderr
    # about this error will not change in the future.
    target_version, current_version = detect_required_terraform_version(terraform.stderr)

    if current_version is not None and target_version is not None:
        sys.stderr.write('. Plan file has been created by terraform v{}, '
                         'but you have v{}\n'.format(target_version, current_version))
        terraform_executable = download_terraform(target_version)
        return convert_terraform_plan_to_json(terraform_plan_file=terraform_plan_file,
                                              terraform_executable=terraform_executable)


    else:
        sys.stderr.write('ERROR: Failed to convert terraform plan file to JSON format via terraform. Here is the error :\n')
        print(terraform.stdout)
        print(terraform.stderr)

        if 'Could not satisfy plugin requirements' in terraform.stderr:
            print('Hint: You can avoid this problem by converting your plan file to a JSON file via running;\n '
                  '\n    # terraform show -json {} > {}.json'
                  '\n\n                          OR'
                  '\n    # terraform init'
                  '\n\n in {} directory and then pass (with -p) {}.json to terraform-compliance'.format(terraform_plan_file,
                                                                                                        terraform_plan_file,
                                                                                                        path,
                                                                                                        terraform_plan_file))

        sys.exit(1)


def get_platform_details():
    system = platform.system().lower()
    arch = "amd64" if platform.machine() == "x86_64" else "386"
    return system, arch


def download_terraform(version):
    system, arch = get_platform_details()
    tmp_dir = tempfile.gettempdir()
    terraform_file = "{}/terraform_{}_{}_{}".format(tmp_dir, version, system, arch)

    if os.path.isfile(terraform_file):
        print('. Using cached {}'.format(terraform_file))
        return terraform_file

    url = 'https://releases.hashicorp.com/terraform/{}/terraform_{}_{}_{}.zip'.format(version, version, system,
                                                                                      arch)
    print('. Downloading terraform v{} from {} ...'.format(version, url))
    urllib.request.urlretrieve(url, '{}.zip'.format(terraform_file))
    print('. Unpacking {}.zip'.format(terraform_file))
    unpack_archive('{}.zip'.format(terraform_file), tmp_dir)
    os.rename('{}/terraform'.format(tmp_dir), terraform_file)
    st = os.stat(terraform_file)
    os.chmod(terraform_file, st.st_mode | stat.S_IEXEC)
    return terraform_file


def detect_required_terraform_version(string):
    match_regex = [
        r'([0-9.]+), but this is ([0-9.]+); plan files cannot be transferred between',
        r'by Terraform v([0-9.]+), which is newer than current v([0-9.]+);'
    ]
    for pattern in match_regex:
        match = re.search(pattern, string, re.DOTALL)

        if match is not None:
            return match.group(1), match.group(2)

    return None, None


# -*- coding: utf-8 -*-

from terraform_compliance.common.helper import (
    Null,
)
from terraform_compliance.common.error_handling import Error
from terraform_compliance.common.exceptions import TerraformComplianceNotImplemented


def its_key_condition_be_value(_step_obj, key, condition, value, stash=Null, depth=0):
    match = _step_obj.context.match
    seek_regex_key_in_dict_values = match.seek_regex_key_in_dict_values

    if condition not in ('must', 'must not'):
        raise TerraformComplianceNotImplemented('This step only accepts "must" and "must not" as a condition.')
    condition = condition == 'must'

    if stash is Null:
        stash = _step_obj.context.stash

    if not stash or stash is Null:
        Error(_step_obj, 'No entities found for this step to process. Check your filtering steps in this scenario.')
        return False

    obj_address = _step_obj.context.name
    if hasattr(_step_obj.context, 'address'):
        obj_address = _step_obj.context.address
    elif hasattr(_step_obj.context, 'addresses'):
        obj_address = ', '.join(_step_obj.context.addresses)

    found_values = []
    for entity in stash:
        if isinstance(entity, dict):
            found_value = seek_regex_key_in_dict_values(entity, key, value)
            if not found_value and condition:
                Error(_step_obj, 'Can not find {} in {} property of {}.'.format(value, key, match.get(entity, 'address', obj_address)))                
                Error(_step_obj, 'Can not find {} in {} property of {}.'.format(value, key, obj_address)) # legacy error message
            elif found_value and not condition:
                Error(_step_obj, 'Found {}({}) in {} property of {}.'.format(value, ', '.join(found_values), key, obj_address))

            found_values.extend(found_value)

        elif isinstance(entity, list):
            dict_entity = {f'not_{str(key)}': entity}
            found_value = seek_regex_key_in_dict_values(dict_entity, key, value)
            if not found_value and condition:
                Error(_step_obj, 'Can not find {} in {} property of {}.'.format(value, key, obj_address))
            elif found_value and not condition:
                Error(_step_obj, 'Found {}({}) in {} property of {}.'.format(value, ', '.join(found_values), key, obj_address))

            found_values.extend(found_value)

        elif isinstance(entity, (str, int, bool)):
            # raise error because you don't have a {key: value}
            if (match.equals(entity, key) or match.equals(entity, value)) and condition:
                Error(_step_obj, 'Value {} found in {} property of {}, but is not in {{key: value}} format.'.format(value, key, obj_address))
            if condition:
                Error(_step_obj, 'Can not find {} in {} property of {}.'.format(value, key, obj_address))


    _step_obj.context.stash = found_values

    return True



import colorful

from radish.hookregistry import after
from radish.extensionregistry import extension
from radish.utils import console_write

from radish.extensions.formatters.gherkin import ConsoleWriter
ConsoleWriter.OPTIONS = []

@extension
class SilentFormatter(ConsoleWriter):
    """
        This class is a simple extension to ConsoleWriter (a.k.a config.formatter == 'gherkin')
    """

    LOAD_IF = staticmethod(lambda config: config.formatter == "silent_formatter")
    LOAD_PRIORITY = 30


    def __init__(self):
        after.each_feature(self.silent_formatter_after_each_feature)

    def silent_formatter_after_each_feature(self, feature):
        """
            Writes failing features to the console

            :param Feature feature: the feature to write to the console
        """
        if not any(scenario.state == 'failed' for scenario in feature.all_scenarios):
            return

        # ConsoleWriter.console_writer_before_each_feature(self, feature)
        self.console_writer_before_each_feature(feature)

        for scenario in feature.all_scenarios:
            self.silent_formatter_after_each_scenario(scenario)
        
        # one newline between final scenario and results summary
        console_write('')


    def silent_formatter_after_each_scenario(self, scenario):
        """
            similar to call gherkin's formatter but with appropriate modifications

            :param Scenario scenario: the scenario which was ran.
        """
        if scenario.state != 'failed':
            return

        ConsoleWriter.console_writer_before_each_scenario(self, scenario)

        msg_header = '\t\t\x1b[1m\x1b[31mFailure\x1b[1m\x1b[37m:\x1b[22m\x1b[39m\x1b[1m\x1b[31m\x1b[22m\x1b[39m\x1b[26m \x1b[31m'
        for step in scenario.all_steps:
            if step.state == step.State.FAILED:
                # hardcoded formatting
                if len(step.context.failure_msg) == 2:
                    step.failure.reason = step.context.failure_msg[1]

                elif len(step.context.failure_msg) > 2:
                    out = '{}\n\t  '.format(step.context.failure_msg[1]) # radish fills the header for the first message
                    messages = step.context.failure_msg[3::2] # remaining messages
                    headers = step.context.failure_msg[2::2] # remaining headers headers
                    out += '\n\t  '.join(['{} {}'.format(headers[i], messages[i]) for i in range(len(messages))])
                    step.failure.reason = out

            self.console_writer_after_each_step(step)

        self.console_writer_after_each_scenario(scenario)




import json
import re
from IPython import embed
from radish import before, after, world
from terraform_compliance.extensions.terraform import TerraformParser
from terraform_compliance.common.helper import Null
from terraform_compliance.extensions.ext_radish_bdd import skip_step

@before.each_feature
def load_terraform_data(feature):
    world.config.terraform = TerraformParser(world.config.user_data['plan_file'])


@before.each_step
def parse_in_step_variables(step):
    # ignore given steps (doesn't support match, would need to do something else for them, also there's no second stash)
    if step.context_class == 'given':
        return

    # if step doesn't have argument_match, it doesn't support in_step_variables
    if step.argument_match is None:
        return

    if step.context.bad_tags:
        return

    groupindex = dict(step.argument_match.match.match.re.groupindex)
    regs = step.argument_match.match.match.regs
    sentence = step.context_sensitive_sentence
    field_map = {field: sentence[regs[i][0]: regs[i][1]] for field, i in groupindex.items()}

    # ignore the regex statements
    if 'search_regex' in field_map:
        del field_map['search_regex']

    in_step_variables = {}
    match = step.context.match
    regex = r'(.*){(.*)}(.*)'

    for field, field_val in field_map.items():
        matches = match.regex_match(regex, str(field_val))
        if matches is None:
            continue
        
        query = matches.group(2).split('.')
        if not query:
            continue # do something here
        
        resource_type = query.pop(0)
        query_result = [resource for resource in step.context.cumulative_stash if match.equals(resource['type'], resource_type)]

        list_indexing_regex = r'\[([0-9]*)(:)?([0-9]*)\]' # accepts []
        for q in query:
            if not query_result:
                break
            
            list_indexing_match = match.regex_match(list_indexing_regex, q)
            if list_indexing_match and q != '[]':
                if not list_indexing_match.group(2):
                    # index from the first group
                    index = int(list_indexing_match.group(1))
                    query_result = [resource[index] for resource in query_result]
                else:
                    # slice
                    temp_query_result = query_result[:]
                    index1 = int(list_indexing_match.group(1)) if list_indexing_match.group(1) else 0
                    index2 = int(list_indexing_match.group(3)) if list_indexing_match.group(3) else len(temp_query_result)
                    temp_query_result = temp_query_result[index1:index2]

                    query_result = [item for resource in temp_query_result for item in resource]
                    
            else:
                query_result = [match.get(resource, q) for resource in query_result if match.contains(resource, q)]
        
        if matches.group(1) or matches.group(3):
            if not all(isinstance(resource, (str, bool, int, float)) for resource in query_result):
                raise TypeError("Improper in step variable usage. Can't mix affixes with non-str in step variabes.")
            
            query_result = [f'{matches.group(1)}{resource}{matches.group(3)}' for resource in query_result]

        in_step_variables[field] = query_result

    step.context.in_step_variables = in_step_variables

@after.each_step
def exclude_resources(step):
    if not hasattr(step.context, 'resources_to_exclude') or not step.context.resources_to_exclude:
        return 

    if step.context_class != 'given' or not step.context.stash:
        return
    
    match = step.context.match
    resources_to_exclude = step.context.resources_to_exclude
    resources = []

    for resource in step.context.stash:
        address = resource.get('address', '')
        if match.regex_match(resources_to_exclude, address) is None:
            resources.append(resource)
    
    step.context.stash = resources

    # if stash is empty, skip
    # note that if the stash was empty from the Given step, this function will not run
    if not step.context.stash:
        skip_step(step, step.context.name)

# debugging step should be always the last hooker
@after.each_step(order=1)
def wait_for_user_input(step):
    if world.config.user_data['debugging_mode_enabled'] == 'False':
        return

    cmd = 'cmd'
    while cmd != '':
        try:
            cmd = input(">> Press enter to continue")
        except EOFError:
            print()
            return
        
        if cmd == 'd':
            embed()
        elif cmd == 's':
            print(json.dumps(step.context.stash, indent=4))
        elif cmd != '':
            print(
                """
Commands
- s: prints stash
- d: opens Interactive Python
- h: prints commands
                """
            )


